#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/nuclear/t2/T2SimpleMetrics.py
# License           : BSD-3-Clause
# Author            : mitchell@nyu.edu
# Date              : 11.05.2020
# Last Modified Date: 08.06.2020
# Last Modified By  : mitchell@nyu.edu

import numpy as np
import os, sys
import numpy as np
from astropy import units as u  # type: ignore
from astropy.coordinates import SkyCoord  # type: ignore
import astropy.time  # type: ignore
import pandas as pd  # type: ignore

from ampel.types import UBson
from ampel.struct.UnitResult import UnitResult
from ampel.abstract.AbsLightCurveT2Unit import AbsLightCurveT2Unit
from ampel.ztf.util.ZTFIdMapper import to_ztf_id
from ampel.view import LightCurve


class T2SimpleMetrics(AbsLightCurveT2Unit):
    """
    Computes simple lightcurve metrics
    """

    def process(self, light_curve) -> UBson | UnitResult:
        """
        Parameters
        ----------
        light_curve: `ampel.view.LightCurve` instance.
        See original docstring & photopoint docstring.
        Note: arguments of '.get_values()' method are strings and are in the following table: https://zwickytransientfacility.github.io/ztf-avro-alert/schema.html

        run_config: 'dict'
        'local_data': plotting folder
        'name': ZTF name of object,

        Returns
        -------
        Dict of two dicts:
        {
        metrics: Dict of simple metrics and their values
        plot_info: Info used for plotting t3

        }
        """
        lc = _lc_to_dict(light_curve)
        name = to_ztf_id(light_curve.get_photopoints()[0]["stock"][0])
        meta_data = {"name": name}
        result = jsonify(compute(meta_data, lc, self.logger))
        return result


def _lc_to_dict(lc):
    # Turns Ampel Lightcurve object into a dict
    keys = (
        "jd",
        "fid",
        "pid",
        "diffmaglim",
        "pdiffimfilename",
        "programpi",
        "programid",
        "candid",
        "isdiffpos",
        "tblid",
        "nid",
        "rcid",
        "field",
        "xpos",
        "ypos",
        "ra",
        "dec",
        "magpsf",
        "sigmapsf",
        "chipsf",
        "magap",
        "sigmagap",
        "distnr",
        "magnr",
        "sigmagnr",
        "chinr",
        "sharpnr",
        "sky",
        "magdiff",
        "fwhm",
        "classtar",
        "mindtoedge",
        "magfromlim",
        "seeratio",
        "aimage",
        "bimage",
        "aimagerat",
        "bimagerat",
        "elong",
        "nneg",
        "nbad",
        "rb",
        "rbversion",
        "drb",
        "drbversion",
        "ssdistnr",
        "ssmagnr",
        "ssnamenr",
        "sumrat",
        "magapbig",
        "sigmagapbig",
        "ranr",
        "decnr",
        "ndehist",
        "ncovhist",
        "jdstarhist",
        "jdendhist",
        "scorr",
        "tooflag",
        "objectidps1",
        "sgmag1",
        "srmag1",
        "simag1",
        "szmag1",
        "sgscore1",
        "distpsnr1",
        "objectidps2",
        "sgmag2",
        "srmag2",
        "simag2",
        "szmag2",
        "sgscore2",
        "distpsnr2",
        "objectidps3",
        "sgmag3",
        "srmag3",
        "simag3",
        "szmag3",
        "sgscore3",
        "distpsnr3",
        "nmtchps",
        "rfid",
        "jdstartref",
        "jdendref",
        "nframesref",
        "dsnrms",
        "ssnrms",
        "dsdiff",
        "magzpsci",
        "magzpsciunc",
        "magzpscirms",
        "nmatches",
        "clrcoeff",
        "clrcounc",
        "zpclrcov",
        "zpmed",
        "clrmed",
        "clrrms",
        "neargaia",
        "neargaiabright",
        "maggaia",
        "maggaiabright",
        "exptime",
    )

    vals = []

    for key in keys:
        vals.append(np.array(lc.get_values(key)))

    return dict(zip(keys, vals))


def wmean(d, ivar):
    """
    >> mean = wmean(d, ivar)
    inverse-variance weighted mean
    """
    d = np.array(d)
    ivar = np.array(ivar)
    return np.sum(d * ivar) / np.sum(ivar)


def jsonify(obj):
    """
    Recursively cast to JSON native types
    """
    if hasattr(obj, "tolist"):
        return obj.tolist()
    elif hasattr(obj, "keys"):
        return {jsonify(k): jsonify(obj[k]) for k in obj.keys()}
    elif hasattr(obj, "__len__") and not isinstance(obj, str):
        return [jsonify(v) for v in obj]
    else:
        return obj


def compute(meta_data, lc, logger, verbose=False):

    # make the rec
    result = {}
    plot_info = {}

    ztfname = meta_data["name"]

    # can be useful to remake offset plots, including ps1_data
    plot_info["plot_fname"] = ztfname + "_offset.pdf"
    plot_info["offset_plot"] = False

    # force remake plot if source is still young
    result["name"] = ztfname

    result["ra"], result["dec"] = lc["ra"], lc["dec"]

    data_dict1 = {
        key: value for key, value in lc.items() if len(value) == len(lc["jd"])
    }
    photoarr0 = pd.DataFrame(data_dict1)

    isdetect = photoarr0["magpsf"] < 99  # should be everything
    idiffspos = photoarr0["isdiffpos"] == "t"

    result["ndetections"] = int(sum(idiffspos & isdetect))
    result["ndetections_negative"] = int(sum((idiffspos == False) & isdetect))

    source_age = astropy.time.Time.now().jd - max(lc["jd"])
    # logging.info('simple_metrics: {0} ({1});  age={2:4.1f}d; ra,dec = {3:0.5f} {4:0.5f}; ndetection={5}'.format(ztfname, str(meta_data['classification']), source_age, result['ra'],result['dec'], result['ndetections'] ))

    # we only work with positive detections
    if result["ndetections"] < 3:
        logger.warn("WARNING. not enough positive detection for this source!")

        return {"metrics": result, "plot_info": plot_info}

    photoarr = photoarr0[idiffspos & isdetect]
    photoarr = photoarr.iloc[np.argsort(photoarr["jd"])]

    # here we fill in the metric
    ipeak = photoarr["magpsf"] == min(
        photoarr["magpsf"]
    )  # not very robust but OK for now
    ilatest = photoarr["jd"] == max(photoarr["jd"])

    result["peak_mag"] = min(photoarr["magpsf"])
    result["latest_mag"] = max(photoarr[ilatest]["magpsf"])
    result["peak_jd"] = min(photoarr[ipeak]["jd"])
    result["latest_jd"] = max(photoarr[ilatest]["jd"])
    result["age"] = astropy.time.Time.now().jd - result["latest_jd"]

    for flt in [1, 2]:

        photoarr_flt = photoarr[photoarr["fid"] == flt]
        flt_name = "g" if flt == 1 else "r" if flt == 2 else None
        if "ranr" in photoarr.columns:
            result["ndetections_{0}".format(flt_name)] = len(photoarr_flt)
        else:
            result["ndetections_{0}".format(flt_name)] = 0

        if result["ndetections_{0}".format(flt_name)] > 2:

            dra = (
                photoarr_flt["ra"] - photoarr_flt["ranr"]
            ) * 3600  # offset to nearest ZTF PSF catalog source
            ddec = (photoarr_flt["dec"] - photoarr_flt["decnr"]) * 3600
            drms = np.clip(np.std(np.append(dra, ddec)), 0.1, 1)
            dstd = np.clip(drms / np.sqrt(len(photoarr_flt)), 0.05, 1)

            # offset_sigma = np.clip(0.27 + 1.66 * (photoarr_flt['sigmamagpsf']-0.1) , 0.1, 10)  # equation from the NedStark paper
            sigma_offset = np.clip(
                0.24 + 0.04 * (photoarr_flt["magpsf"] - 20), 0.1, 1
            )  # equation from the NedStark paper
            snr_weight = 1 / sigma_offset**2

            if "magnr" in photoarr.columns:
                sigma_offset_nr = np.clip(
                    0.24 + 0.04 * (np.median(photoarr_flt["magnr"]) - 20), 0.1, 1
                )  # / np.sqrt(5) # assume x images make a ref frame
            else:
                sigma_offset_nr = 0.05

            dunc_theo = np.sqrt(
                1 / (sum(1 / sigma_offset**2)) + sigma_offset_nr**2
            )  # can be too optimistic if large number of observations
            dunc_rms = np.sqrt(
                dstd**2 + sigma_offset_nr**2
            )  # can be too optimisiric if small number of observations
            dunc = np.max([dunc_rms, dunc_theo])

            weighted_dra = wmean(dra, snr_weight)
            weighted_dec = wmean(ddec, snr_weight)
            logger.info(
                "weighted (dra,drec) {0:0.5f} {1:0.5f}".format(
                    weighted_dra, weighted_dec
                )
            )
            offset_weighted = np.sqrt(weighted_dra**2 + weighted_dec**2)

            result["offset_rms_{0}".format(flt_name)] = np.std(np.append(dra, ddec))
            result["offset_unc_{0}".format(flt_name)] = dunc
            result["offset_med_{0}".format(flt_name)] = np.sqrt(
                np.median(dra) ** 2 + np.median(ddec) ** 2
            )
            result["offset_weighted_{0}".format(flt_name)] = offset_weighted
            result["offset_sig_{0}".format(flt_name)] = offset_weighted / dunc

            logger.info(
                str(flt_name)
                + ": "
                + "simple_metrics: distance to ref (median, weighted, significance): {0:0.3f} {1:0.3f} {2:0.3f}".format(
                    result["offset_med_{0}".format(flt_name)],
                    result["offset_weighted_{0}".format(flt_name)],
                    result["offset_sig_{0}".format(flt_name)],
                )
            )
            logger.info(
                str(flt_name)
                + ": "
                + "simple_metrics: distance to ref (rms, unc_rm, unc_theo) : {0:0.3f} {1:0.3f} {2:0.3f}".format(
                    result["offset_rms_{0}".format(flt_name)], dunc_rms, dunc_theo
                )
            )

        else:
            result["offset_sig_{0}".format(flt_name)] = -999

    # select which offset measurement to use based on rm
    if (result["ndetections_g"] > 2) and (result["ndetections_r"] > 2):
        if result["offset_unc_g"] < result["offset_unc_r"]:
            result["offset_sig"] = result["offset_sig_g"]
        else:
            result["offset_sig"] = result["offset_sig_r"]
    elif result["ndetections_r"] > 2:
        result["offset_sig"] = result["offset_sig_r"]
    elif result["ndetections_g"] > 2:
        result["offset_sig"] = result["offset_sig_g"]

    result["lastest_obs_is_detection"] = int(
        max(photoarr0["jd"]) == result["latest_jd"]
    )
    result["latest_maglim"] = min(
        photoarr[ilatest]["diffmaglim"]
    )  # assumed limmag meant diffmaglim

    # by default assume we have no detection in ref image
    result["ref_mag_r"], result["ref_mag_g"] = 23, 23

    # get ref magnitude if detected (note cut on distnr is needed to avoid matches to nearby unrelated sources for non-detections in a single band)
    if "magnr" in photoarr.columns:

        for flt in (1, 2):

            flt_name = "g" if flt == 1 else "r" if flt == 2 else None
            ii = (photoarr["fid"] == flt) & (photoarr["distnr"] < 2)
            if sum(ii):
                result["ref_mag_" + flt_name] = np.median(photoarr[ii]["magnr"])
    else:
        logger.warn(f"warning: magnr not found in photoarr: {photoarr.columns}")

    # now compute magnitude difference
    for flt in (1, 2):

        flt_name = "g" if flt == 1 else "r" if flt == 2 else None
        ii = photoarr["fid"] == flt

        if sum(ii):
            ip = photoarr["magpsf"][ii] == min(photoarr[ii]["magpsf"])
            il = photoarr[ii]["jd"] == max(photoarr[ii]["jd"])

            result["peak_mag_" + flt_name] = (photoarr["magpsf"][ii][ip]).values[0]
            result["latest_mag_" + flt_name] = (photoarr["magpsf"][ii][il]).values[0]
            result["peak_diff_mag_" + flt_name] = (
                (photoarr["magpsf"][ii] - result["ref_mag_" + flt_name])[ip]
            ).values[0]
            result["latest_diff_mag_" + flt_name] = (
                (photoarr["magpsf"][ii] - result["ref_mag_" + flt_name])[il]
            ).values[0]
            logger.info(
                str(flt_name)
                + ": "
                + "(ref_mg, peak_diff_mag)  = ({0:0.2f}, {1:0.2f})".format(
                    result["ref_mag_" + flt_name], result["peak_diff_mag_" + flt_name]
                )
            )
        else:
            result["peak_mag_" + flt_name] = 999
            result["latest_mag_" + flt_name] = 999
            result["peak_diff_mag_" + flt_name] = 999
            result["latest_diff_mag_" + flt_name] = 999

    # finally, store the largest flux increase between the two bands
    result["ref_mag"] = min(result["ref_mag_r"], result["ref_mag_g"])
    result["latest_diff_mag"] = min(
        result["latest_diff_mag_r"], result["latest_diff_mag_g"]
    )
    result["peak_diff_mag"] = min(result["peak_diff_mag_r"], result["peak_diff_mag_g"])

    if "ranr" in photoarr.columns:
        plot_info["offset_plot"] = True
        plot_info["dra"] = (
            photoarr["ra"] - photoarr["ranr"]
        ) * 3600  # offset to nearest ZTF PSF catalog source
        plot_info["ddec"] = (photoarr["dec"] - photoarr["decnr"]) * 3600

        sigma_offset = np.clip(
            0.24 + 0.04 * (photoarr["magpsf"] - 20), 0.1, 1
        )  # equation from the NedStark paper
        plot_info["snr_weight"] = 1 / sigma_offset**2

        plot_info["ig"] = photoarr["fid"] == 1

        plot_info["ir"] = photoarr["fid"] == 2

    return {"metrics": result, "plot_info": plot_info}



"""
least squares fitting on (ZTF) data with observations in two bands
we fit for the peak, using a Gaussian rise and an exponential decay
we measure the mean color pre-peak, and allow for a color-change post-peak

goal is photometric typing of alerts into:

 - "SN" or "Ia" : based on rise time 
 - "AGN"		: slow rise or fading
 - "CV"			: fast
 - "TDE"		: not the above and constant color or blue

"""

import numpy as np

log10, log, exp, sqrt = np.log10, np.log, np.exp, np.sqrt
from scipy.optimize import leastsq
import astropy.stats


# flux/mag conversion without zeropoint (only used to force fit in mag space)
def f2m(flux):
    return -2.5 * log10(flux)


def m2f(mag):
    return 10 ** (-0.4 * (mag))


def line(p, x):
    return p[0] + p[1] * x


def gauss(x, sigma):
    return np.exp(-0.5 * x**2 / (sigma**2))


minfs, maxfs = log(1e-9), log(1e9)  # max fit slope for exponential/gaussian


def broken_gauss(p, x, dt_fix=False, fid=None):

    p = list(p)
    # check if we want to return only a fading/rising function
    if dt_fix is not False:
        x_peak = dt_fix

        if dt_fix > 0:  # rising only
            a1 = m2f(p[0])
            p[1] = np.clip(p[1], minfs, maxfs)
            b1 = exp(p[1])
            return a1 * gauss(x - x_peak, b1)
        else:  # fading only
            a2 = m2f(p[0])
            p[1] = np.clip(p[1], minfs, maxfs)
            b2 = exp(p[1])
            return a2 * np.exp(-(x - x_peak) / b2)

    # gaussian rise
    x_peak = np.clip(p[0], 0, max(x))
    a1 = m2f(p[1])
    p[2] = np.clip(p[2], minfs, maxfs)
    p[3] = np.clip(p[3], minfs, maxfs)
    b1 = exp(p[2])
    leftside = a1 * gauss(x - x_peak, b1)

    # exponential decay
    a2 = a1 * gauss(0, b1)
    b2 = exp(p[3])
    rightside = a2 * np.exp(-(x - x_peak) / b2)

    leftside[x > x_peak] = 0
    rightside[x <= x_peak] = 0

    return leftside + rightside


minmaxcs = 0.1  # min/max colorslope
mincolor, maxcolor = -1, +3


def color_func(x, x_peak, mean_color_in, slope_in):

    mean_color = np.clip(mean_color_in, mincolor, maxcolor)
    slope = np.clip(slope_in, -minmaxcs, minmaxcs)
    color = mean_color + (x - x_peak) * slope
    # ileft = x<=x_peak
    # if sum(ileft) and (x_peak>0):
    # 	color[ileft] = mean_color
    return color


def broken_gauss_twocolor(
    p, x, dt_fix=False, fid=None, flt1="g", flt2="r", x_peakc=None
):

    p = list(p)
    both = broken_gauss(p, x, dt_fix=dt_fix)

    if dt_fix is not False:
        if not x_peakc:
            x_peakc = np.median(x)  # new in 2019 (was 0)
        p[2] = np.clip(p[2], mincolor, maxcolor)
        # p[3] = np.clip(p[3], -minmaxcs, minmaxcs)
        if dt_fix > 0:
            p_col = p[2], 0
        else:
            p_col = p[2], p[3]
    else:
        if not x_peakc:
            x_peakc = np.median(x)  # new in 2019 np.clip(p[0], 0, max(x))
        p[4] = np.clip(p[4], mincolor, maxcolor)
        p[5] = np.clip(p[5], -minmaxcs, minmaxcs)
        p_col = p[4], p[5]

    iflt1 = fid == flt1
    iflt2 = fid == flt2

    color = color_func(x, x_peakc, p_col[0], p_col[1])

    # this can be tricky, when one band has very little data thing can go wrong
    both[iflt1] *= 10 ** (-0.4 * color[iflt1] / 2.0)
    both[iflt2] *= 10 ** (+0.4 * color[iflt2] / 2.0)

    return both


def res(p, x, y, yerr, model_func, dt_fix=False, fid=None, islim=None):

    model_flux = model_func(p, x, dt_fix=dt_fix, fid=fid)

    chi = (model_flux - y) / yerr
    chi = np.clip(chi, -1e5, 1e5)

    # do a rough prior on rise time, increase chi2 if larger/smaller than 100/0.01

    if (model_func is broken_gauss) or (model_func is broken_gauss_twocolor):

        p_col = 0, 0
        # check if we we have a fixed time of peak
        if dt_fix is False:
            logb1 = p[2]
            logb2 = p[3]
            if model_func is broken_gauss_twocolor:
                p_col = p[4], p[5]
        else:
            if dt_fix > 0:  # rising only
                logb1 = p[1]
                logb2 = 0
                if model_func is broken_gauss_twocolor:
                    p_col = p[2], 0
            else:  # fading only
                logb1 = 0
                logb2 = p[1]
                if model_func is broken_gauss_twocolor:
                    p_col = p[2], p[3]

        # log(100) = 4.60517
        rfmax = log(100)
        if abs(logb1) > rfmax:
            chi += (abs(logb1) - rfmax) / len(x)
        if abs(logb2) > 4.605:
            chi += (abs(logb2) - rfmax) / len(x)

        # weak push for constant color (helpful if limited data available?)
        chi += (p_col[1] - 0) / 0.05 / len(x)

    return chi


def flex_fit_wclip(
    dtime,
    flux,
    flux_err,
    fid,
    logger,
    isdetect=None,
    filters=["g", "r"],
    ax=None,
    verbose=True,
    niter=0,
):
    """
    run flex fit a few time for clipping
    also pick the preferred band for reporting the results
    """

    lc_data = {}
    plot_info = {}
    result, lsq, flux_diff = flex_fit(
        dtime,
        flux,
        flux_err,
        fid,
        logger,
        isdetect=isdetect,
        filters=["g", "r"],
        ax=ax,
        verbose=True,
    )
    # print('dtime:', dtime)
    (
        lc_data["dtime"],
        lc_data["flux"],
        lc_data["flux_err"],
        lc_data["fid"],
        lc_data["isdetect"],
    ) = (dtime, flux, flux_err, fid, isdetect)
    plot_info["lsq_out"] = lsq
    plot_info["dt_fix"] = result["dt_fix"]
    outliers = (abs(flux_diff / flux_err) > 7) * isdetect
    logger.info("flex_fit_wclip: # of outliers {0}".format(sum(outliers)))

    # ---
    # pick what band(s) to use store the final results
    final_result = {}
    flt_pick = None
    if ("color" in lsq) and (("r" in lsq) or ("g" in lsq)):
        flt_pick = "color"
    elif "r" in lsq:
        flt_pick = "r"
    elif "g" in lsq:
        flt_pick = "g"
    if flt_pick is not None:
        for item in result[flt_pick]:
            final_result[item] = result[flt_pick][item]
            final_result["band"] = flt_pick
    else:
        logger.info("flex_fit_wclip: no band yielded a succesful fit")

    if (sum(outliers) == 0) or niter > 5:
        plot_info["dtime_out"], plot_info["flux_out"] = None, None
        return final_result, lc_data, plot_info
    else:
        # print (flux_diff/flux_err)
        # print (flux_diff)
        logger.info(
            "flex_fit_wclip: running again after outlier rejection, attempts left={0}".format(
                5 - niter
            )
        )

        # allow one rejection
        maxout = max(abs(flux_diff[isdetect] / flux_err[isdetect]))
        outliers = abs(flux_diff / flux_err) == maxout
        logger.info(f"outliers with flux: {flux[outliers]}")

        plot_info["dtime_out"] = dtime[outliers]
        plot_info["flux_out"] = flux[outliers]

        dtime = dtime[outliers == False]
        flux = flux[outliers == False]
        flux_err = flux_err[outliers == False]
        fid = fid[outliers == False]
        isdetect = isdetect[outliers == False]

        if len(flux[isdetect]) < 4:
            logger.info("flex_fit_wclip: not enough data left, keeping original result")
            return final_result, lc_data, plot_info

        # loop continues
        return flex_fit_wclip(
            dtime,
            flux,
            flux_err,
            fid,
            logger,
            isdetect=isdetect,
            niter=niter + 1,
            filters=["g", "r"],
            ax=ax,
            verbose=verbose,
        )


def flex_fit(
    dtime,
    flux,
    flux_err,
    fid,
    logger,
    isdetect=None,
    filters=["g", "r"],
    ax=None,
    verbose=True,
):
    """
    function to fit a light curve with one or two bands with two Gaussian and get some color info
    returning scipy.optimize.leastq output dicts for each band and for the fit to both bands

    observations that are upper limits should have zero flux and be indicated with the isdetect=[] input (boolians array)
    """

    from .flex_metrics import flux2mag, mag2flux

    # unless otherwise indicated, asume all data is a detection
    if isdetect is None:
        isdetect = np.repeat(True, len(dtime))

    # output dicts
    dt_fix = {}
    result = {k: {} for k in filters + ["color"]}
    lsq_out = {}

    # do initial guess of parameters
    flux_peak_guess = flux[isdetect][
        np.argsort(flux[isdetect])[-1]
    ]  # pick brightest point as peak
    t_peak_guess = dtime[isdetect][np.argmax(flux[isdetect])]
    islim = isdetect == False

    flux_diff = np.zeros(len(flux))

    # irise = dtime<t_peak_guess0:
    # if sum(irise):
    sigma_rise_guess = np.clip(t_peak_guess - min(dtime[isdetect]), 1, 20)
    ipost = dtime > t_peak_guess
    if sum(ipost * isdetect):
        sigma_fade_guess = np.clip(
            np.interp(
                flux_peak_guess / 2.0,
                flux[ipost * isdetect][::-1],
                dtime[ipost * isdetect][::-1],
            )
            - t_peak_guess,
            1,
            40,
        )
    else:
        sigma_fade_guess = 20

    if verbose:
        logger.info(
            "guessed time of peak (wrt first obs) {0:0.2f}".format(t_peak_guess)
        )
        logger.info(
            "guessed peak (flux, mag)  {0:0.2f} {1:0.2f}".format(
                flux_peak_guess, flux2mag(flux_peak_guess)
            )
        )
        logger.info(
            "guessed sigma (rise, fade)  {0:0.2f} {1:0.2f}".format(
                sigma_rise_guess, sigma_fade_guess
            )
        )

    # do single band fit for the two filter in fid
    for k in filters:

        iflt = fid == k
        if verbose:
            logger.info(str(k) + " " + str(sum(iflt * isdetect)) + " " + str(sum(iflt)))

        if (sum(iflt * isdetect) > 2) and (sum(iflt) > 3):

            # do the least-square fit
            p0 = [
                t_peak_guess,
                f2m(flux_peak_guess),
                log(sigma_rise_guess),
                log(sigma_fade_guess),
            ]
            dt_fix[k] = False
            lsq_out[k] = leastsq(
                res,
                p0,
                (dtime[iflt], flux[iflt], flux_err[iflt], broken_gauss, dt_fix[k]),
                full_output=True,
            )

            result[k]["dtime_peak"] = lsq_out[k][0][0]
            result[k]["mag_peak"] = flux2mag(m2f(lsq_out[k][0][1]))
            result[k]["sigma_rise"] = exp(lsq_out[k][0][2])
            result[k]["sigma_fade"] = exp(lsq_out[k][0][3])

            if lsq_out[k][1] is not None:
                result[k]["e_dtime_peak"] = sqrt(lsq_out[k][1][0, 0])
                result[k]["e_mag_peak"] = sqrt(lsq_out[k][1][1, 1])
                result[k]["e_sigma_rise"] = (
                    sqrt(lsq_out[k][1][2, 2]) * result[k]["sigma_rise"]
                )
                result[k]["e_sigma_fade"] = (
                    sqrt(lsq_out[k][1][3, 3]) * result[k]["sigma_fade"]
                )

                if verbose:
                    ss = [
                        "{0:0.2f} ({1:0.2f}) ".format(result[k][x], result[k]["e_" + x])
                        for x in ("dtime_peak", "mag_peak", "sigma_rise", "sigma_fade")
                    ]
                    logger.info(f"rise/fade fit: {ss}")

            # if we dont get a decent fit with errorbars,
            # check that the source is not only risng/fading
            else:

                if verbose:
                    ss = [
                        "{0:0.2f} ".format(result[k][x])
                        for x in ("dtime_peak", "mag_peak", "sigma_rise", "sigma_fade")
                    ]
                    logger.info(f"rise/fade fit: {ss}")

                if lsq_out[k][0][0] > (max(dtime[iflt]) - 3):
                    dt_fix[k] = max(dtime[isdetect])
                    p0 = [f2m(flux_peak_guess), log(sigma_rise_guess)]
                elif lsq_out[k][0][0] < (min(dtime[iflt]) + 3):
                    dt_fix[k] = 0
                    p0 = [f2m(flux_peak_guess), log(sigma_fade_guess)]

            if dt_fix[k] is not False:
                if verbose:
                    logger.info(
                        "retrying fit without break, using x_peak fixed at {0:0.2f}".format(
                            dt_fix[k]
                        )
                    )
                lsq_out[k] = leastsq(
                    res,
                    p0,
                    (dtime[iflt], flux[iflt], flux_err[iflt], broken_gauss, dt_fix[k]),
                    full_output=True,
                )

                if verbose:
                    logger.info(
                        "dt_fix fit: {0:0.2f} {1:0.2f}".format(
                            lsq_out[k][0][0], lsq_out[k][0][1]
                        )
                    )

                # overwrite results with single fit
                result[k]["dtime_peak"] = dt_fix[k]
                result[k]["e_dtime_peak"] = -1.0
                result[k]["mag_peak"] = flux2mag(m2f(lsq_out[k][0][0]))
                if lsq_out[k][1] is not None:
                    result[k]["e_mag_peak"] = sqrt(lsq_out[k][1][0, 0])

                if dt_fix[k] > 0:
                    result[k]["sigma_rise"] = exp(lsq_out[k][0][1])
                    result[k]["sigma_fade"] = 0
                    if lsq_out[k][1] is not None:
                        result[k]["e_sigma_rise"] = (
                            sqrt(lsq_out[k][1][1, 1]) * result[k]["sigma_rise"]
                        )
                else:
                    result[k]["sigma_fade"] = exp(lsq_out[k][0][1])
                    result[k]["sigma_rise"] = 0
                    if lsq_out[k][1] is not None:
                        result[k]["e_sigma_fade"] = (
                            sqrt(lsq_out[k][1][1, 1]) * result[k]["sigma_fade"]
                        )

            result[k]["n_rise"] = int(
                sum(dtime[isdetect * iflt] < result[k]["dtime_peak"])
            )
            result[k]["n_fade"] = int(
                sum(dtime[isdetect * iflt] > result[k]["dtime_peak"])
            )

            flux_model = broken_gauss(lsq_out[k][0], dtime[iflt], dt_fix=dt_fix[k])
            flux_diff[iflt] = flux_model - flux[iflt]

            result[k]["mad"] = astropy.stats.median_absolute_deviation(
                flux_diff[isdetect]
            ) / np.median(flux[isdetect * iflt])
            result[k]["rms"] = np.std(flux_diff[isdetect]) / np.mean(
                flux[isdetect * iflt]
            )
            result[k]["chi2"] = sum(
                (flux_diff[iflt * isdetect] / flux_err[iflt * isdetect]) ** 2
            ) / (sum(isdetect) - len(p0))
            logger.info(
                "chi2={0:0.2f}; RMS={1:0.2f}, MAD={2:0.2f}".format(
                    result[k]["chi2"], result[k]["rms"], result[k]["mad"]
                )
            )

    # ----------
    # fit both colors at the same time
    iflt_0 = fid == filters[0]
    iflt_1 = fid == filters[1]

    # at least 6 detection are needed given the number of free parameters
    # plus we require at least two detections in a given band,
    # otherwise the fit has too much freedom to go wild due to the color evolution term
    if (
        sum(isdetect) > 5
        and (sum(isdetect * iflt_0) > 1)
        and (sum(isdetect * iflt_1) > 1)
    ):

        if verbose:
            logger.info(f"two color {sum(isdetect)}  {len(dtime)}")

        dt_fix["color"] = False
        p0 = [
            t_peak_guess,
            f2m(flux_peak_guess),
            log(sigma_rise_guess),
            log(sigma_fade_guess),
            -0.1,
            0.0,
        ]
        lsq_out["color"] = leastsq(
            res,
            p0,
            (dtime, flux, flux_err, broken_gauss_twocolor, dt_fix["color"], fid),
            full_output=True,
        )  # , maxfev=int(1e6), xtol=1e-99, ftol=1e-99)

        # store results
        result["color"]["dtime_peak"] = lsq_out["color"][0][0]
        result["color"]["mag_peak"] = flux2mag(m2f(lsq_out["color"][0][1]))
        result["color"]["sigma_rise"] = exp(lsq_out["color"][0][2])
        result["color"]["sigma_fade"] = exp(lsq_out["color"][0][3])
        result["color"]["mean_color"] = lsq_out["color"][0][4]
        result["color"]["color_slope"] = lsq_out["color"][0][5]

        if lsq_out["color"][1] is not None:
            result["color"]["e_dtime_peak"] = sqrt(lsq_out["color"][1][0, 0])
            result["color"]["e_mag_peak"] = sqrt(lsq_out["color"][1][1, 1])
            result["color"]["e_sigma_rise"] = (
                sqrt(lsq_out["color"][1][2, 2]) * result["color"]["sigma_rise"]
            )
            result["color"]["e_sigma_fade"] = (
                sqrt(lsq_out["color"][1][3, 3]) * result["color"]["sigma_fade"]
            )
            result["color"]["e_mean_color"] = sqrt(lsq_out["color"][1][4, 4])
            result["color"]["e_color_slope"] = sqrt(lsq_out["color"][1][5, 5])

            if verbose:
                ss = [
                    "{0:0.2f} ({1:0.2f}) ".format(
                        result["color"][x], result["color"]["e_" + x]
                    )
                    for x in (
                        "dtime_peak",
                        "mag_peak",
                        "sigma_rise",
                        "sigma_fade",
                        "mean_color",
                        "color_slope",
                    )
                ]
                logger.info(f"rise/fade fit: {ss}")

        # again, if we dont get a decent fit with errorbars,
        # check that the source is not only rising/fading
        else:

            if verbose:
                ss = [
                    "{0:0.2f}".format(result["color"][x])
                    for x in (
                        "dtime_peak",
                        "mag_peak",
                        "sigma_rise",
                        "sigma_fade",
                        "mean_color",
                        "color_slope",
                    )
                ]
                logger.info(f"rise/fade fit: {ss}")

            if lsq_out["color"][0][0] > (max(dtime) - 3):
                dt_fix["color"] = max(dtime[isdetect])
                p0 = [f2m(flux_peak_guess), log(sigma_rise_guess), 0, -0.02]
            elif lsq_out["color"][0][0] < (min(dtime) + 3):
                dt_fix["color"] = 0
                p0 = [f2m(flux_peak_guess), log(sigma_fade_guess), 0, -0.02]

        if dt_fix["color"] is not False:

            if verbose:
                logger.info(
                    "retrying fit without break, using x_peak fixed at {0:0.2f} ".format(
                        dt_fix["color"]
                    )
                )

            if dt_fix["color"] > 0:
                if verbose:
                    logger.info("no allowing color evolution")
                p0 = [f2m(flux_peak_guess), log(sigma_rise_guess), -0.1]
            else:
                p0 = [f2m(flux_peak_guess), log(sigma_rise_guess), -0.1, 0]

            lsq_out["color"] = leastsq(
                res,
                p0,
                (dtime, flux, flux_err, broken_gauss_twocolor, dt_fix["color"], fid),
                full_output=True,
            )

            if verbose:
                logger.info(
                    "two-color dt_fix: m_peak={0:0.2f} rise={1:0.2f}     | color={2:0.3f}".format(
                        flux2mag(m2f(lsq_out["color"][0][0])),
                        exp(lsq_out["color"][0][1]),
                        lsq_out["color"][0][2],
                    )
                )

            # overwrite results with single shape two color fit
            result["color"]["dtime_peak"] = dt_fix["color"]
            result["color"]["e_dtime_peak"] = -1
            result["color"]["mag_peak"] = flux2mag(m2f(lsq_out["color"][0][0]))
            result["color"]["mean_color"] = lsq_out["color"][0][2]
            if dt_fix["color"] > 0:
                result["color"]["color_slope"] = 0
            else:
                result["color"]["color_slope"] = lsq_out["color"][0][3]

            if lsq_out["color"][1] is not None:
                result["color"]["e_mag_peak"] = sqrt(lsq_out["color"][1][1, 1])
                result["color"]["e_mean_color"] = sqrt(lsq_out["color"][1][2, 2])
                if dt_fix["color"] > 0:
                    result["color"]["e_color_slope"] = -999
                else:
                    result["color"]["e_color_slope"] = sqrt(lsq_out["color"][1][3, 3])

            if dt_fix["color"] > 0:
                result["color"]["sigma_rise"] = exp(lsq_out["color"][0][1])
                result["color"]["sigma_fade"] = 0
                if lsq_out["color"][1] is not None:
                    result["color"]["e_sigma_rise"] = (
                        sqrt(lsq_out["color"][1][1, 1]) * result["color"]["sigma_rise"]
                    )
            else:
                result["color"]["sigma_fade"] = exp(lsq_out["color"][0][1])
                result["color"]["sigma_rise"] = 0
                if lsq_out["color"][1] is not None:
                    result["color"]["e_sigma_fade"] = (
                        sqrt(lsq_out["color"][1][1, 1]) * result["color"]["sigma_fade"]
                    )

        flux_model = broken_gauss_twocolor(
            lsq_out["color"][0], dtime, dt_fix=dt_fix["color"], fid=fid
        )
        flux_diff = flux_model - flux

        result["color"]["mad"] = astropy.stats.median_absolute_deviation(
            flux_diff[isdetect]
        ) / np.median(flux[isdetect])
        result["color"]["rms"] = np.std(flux_diff[isdetect]) / np.mean(flux[isdetect])
        result["color"]["chi2"] = sum(
            (flux_diff[isdetect] / flux_err[isdetect]) ** 2
        ) / (sum(isdetect) - len(p0))
        logger.info(
            "chi2={0:0.2f}; RMS={1:0.2f}, MAD={2:0.2f}".format(
                result["color"]["chi2"], result["color"]["rms"], result["color"]["mad"]
            )
        )

        result["color"]["n_rise"] = int(
            sum(dtime[isdetect] <= result["color"]["dtime_peak"])
        )
        result["color"]["n_fade"] = int(
            sum(dtime[isdetect] >= result["color"]["dtime_peak"])
        )

    result["dt_fix"] = dt_fix

    return result, lsq_out, flux_diff


# make some higher-level statements about the light curve
def flex_class(result):

    photoclass = ""
    chat_bot = ""

    # check that we have any results (as selected by flex_fit_wclip)
    if not "band" in result:
        return "", "no fit"

    # print color information
    if (result["band"] == "color") and (
        (result["n_rise"] > 2) or (result["n_fade"] > 2)
    ):

        if "e_mean_color" in result:
            if abs(result["e_mean_color"]) < 0.1:
                chat_bot += "<g-r>={0:0.2f}+-{1:0.2f}, ".format(
                    result["mean_color"], np.clip(result["e_mean_color"], 0, 999)
                )

    # print info if we are on the rise
    if result["n_rise"] > 1:
        if result["sigma_rise"] < 300:
            if result["n_fade"] == 0:
                chat_bot += "rising, "
            elif result["n_fade"] < 2:
                chat_bot += "at/near peak, "

            # check slope of rise time for SNe or CV like behaviour
            if not ("e_sigma_rise" in result):
                result["e_sigma_rise"] = 999  # fix for printing

            if not ("e_sigma_fade" in result):
                result["e_sigma_fade"] = 999

            # slow rise time is quite rare, can be AGN, TDE, SN II
            if result["sigma_rise"] > 10:
                photoclass = "TDE?"

                chat_bot += "slow rise time (sigma={0:0.1f}+-{1:0.1f} d), ".format(
                    result["sigma_rise"], np.clip(result["e_sigma_rise"], 0, 999)
                )

                # if also blue this could be a TDE
                if (result["band"] == "color") and (result["mean_color"] < 0.0):
                    photoclass = "TDE"
            elif result["sigma_rise"] > 4:
                chat_bot += "SN-like rise time (sigma={0:0.1f}+-{1:0.1f} d), ".format(
                    result["sigma_rise"], np.clip(result["e_sigma_rise"], 0, 999)
                )
            # 	photoclass = 'SN?'

        else:
            chat_bot += "very slow rise, "
            if (result["n_fade"] > 2) and (result["sigma_fade"] > 500):
                photoclass = "AGN?"

    if (
        (result["n_rise"] > 1)
        and (result["sigma_rise"] < 2)
        and ("e_sigma_rise" in result)
    ):
        chat_bot += "warning, steep increase (sigma={0:0.1f}+-{1:0.1f}), ".format(
            result["sigma_rise"], np.clip(result["e_sigma_rise"], 0, 999)
        )
        photoclass = "CV?"

    if (result["n_rise"] < 2) and (result["sigma_fade"] < 500):
        chat_bot += "missed peak(?), "

    # print info for fading or constant sources
    if (result["n_fade"] > 2) and ("e_sigma_fade" in result):

        if result["sigma_fade"] < 1000:

            chat_bot += "fading (tau={0:0.1f} +- {1:0.1f} d), ".format(
                result["sigma_fade"], np.clip(result["e_sigma_fade"], 0, 999)
            )

        elif result["n_fade"] > 3:

            chat_bot += "very long fading time, "

            # if fading time is actually measured properly, do classification
            if "e_sigma_fade" in result:
                if (result["sigma_fade"] / result["e_sigma_fade"]) > 2:
                    if not photoclass:
                        photoclass = "AGN?"
                elif "color_slope" in result:
                    if abs(result["color_slope"]) < 0.015:
                        photoclass = "not SN?"

    # do checks on color and slope
    if (result["band"] == "color") and ("e_mean_color" in result):

        # add label if blue
        if (result["mean_color"] < -0.2) and (abs(result["e_mean_color"]) < 0.1):
            chat_bot += "blue! ({0:0.1f}+-{1:0.1f}), ".format(
                result["mean_color"], np.clip(result["e_mean_color"], 0, 999)
            )

        # type as TDE if constant color and blue (since May 2019 this now includes ~constant post-peak flux, plus cut on e_color_slope 0.015)
        if result["e_color_slope"] < 0.015:

            if result["color_slope"] < -0.015:
                chat_bot += "getting more blue?! ({0:0.3f}+-{1:0.3f}), ".format(
                    result["color_slope"], np.clip(result["e_color_slope"], 0, 999)
                )
                if (result["mean_color"] < 0) and (result["sigma_fade"] > 5):
                    photoclass = "TDE"
                elif result["mean_color"] < 0.1:
                    photoclass = "TDE?"

            elif abs(result["color_slope"]) < 0.015:
                if result["e_color_slope"] > 0:
                    chat_bot += "near-constant color ({0:0.3f}+-{1:0.3f}) ".format(
                        result["color_slope"], np.clip(result["e_color_slope"], 0, 999)
                    )
                if (result["mean_color"] < 0) and (result["sigma_fade"] > 5):
                    photoclass = "TDE"
                elif result["mean_color"] < 0.1:
                    photoclass = "TDE?"

            elif result["color_slope"] < 0.020:
                chat_bot += "cooling? ({0:0.3f}+-{1:0.3f}), ".format(
                    result["color_slope"], np.clip(result["e_color_slope"], 0, 999)
                )
                if (result["mean_color"] < -0.1) and (result["sigma_fade"] > 5):
                    photoclass = "TDE"
                elif result["mean_color"] < 0.0:
                    photoclass = "TDE?"

            elif (result["color_slope"] >= 0.020) and (result["sigma_fade"] < 100):
                chat_bot += "cooling ({0:0.3f}+-{1:0.3f}), ".format(
                    result["color_slope"], np.clip(result["e_color_slope"], 0, 999)
                )
                if result["color_slope"] != minmaxcs:
                    photoclass = "SN"

    # check fast fading
    if (result["n_fade"] > 2) and (result["sigma_fade"] < 5):
        chat_bot += "warning fast fading, "
        photoclass = "fast/weird"

    if not photoclass:
        photoclass = "unknown"

    if result["chi2"] > 10:
        chat_bot += "warning, poor fit (chi2/dof={0:0.1f})".format(result["chi2"])
        # photoclass +='_badfit'

    return chat_bot.strip(", "), photoclass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/nuclear/t3/T3PlotNeoWISE.py
# License           : BSD-3-Clause
# Author            : mitchell@nyu.edu
# Date              : 08.06.2020
# Last Modified Date: 29.11.2022
# Last Modified By  : simeon.reusch@desy.de

import io, json, warnings
from collections.abc import Generator
from typing import Any, Optional, Union

import astropy.io.ascii  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
import numpy as np
from pydantic import BaseModel
import requests
import tempfile

from ampel.types import UBson, T3Send
from ampel.view.TransientView import TransientView
from ampel.struct.T3Store import T3Store
from ampel.struct.UnitResult import UnitResult
from ampel.abstract.AbsPhotoT3Unit import AbsPhotoT3Unit
from ampel.nuclear.t3.dropboxIO import DropboxUnit
from ampel.secret.NamedSecret import NamedSecret

warnings.filterwarnings("ignore", category=RuntimeWarning)


class T3PlotNeoWISE(DropboxUnit):

    # dropbox_token: None | NamedSecret[str]
    apply_qcuts: bool  #: whether to apply quality cuts to data based on n detections
    plot_allWISE: bool  #: whether or not to plot allWISE data
    verbose: bool = False

    def post_init(self):
        # super().post_init()
        self.save_location = self.base_location + "/alerts"
        if "alerts" not in self.get_files(self.base_location):
            self.create_folder(self.save_location)

    def jsonify(self, obj):
        """
        Recursively cast to JSON native types
        """
        if hasattr(obj, "tolist"):
            return obj.tolist()
        elif hasattr(obj, "keys"):
            return {self.jsonify(k): self.jsonify(obj[k]) for k in obj.keys()}
        elif hasattr(obj, "__len__") and not isinstance(obj, str):
            return [self.jsonify(v) for v in obj]
        else:
            return obj
    
    def get_data_log(self) -> dict[str,Any]:
        try:
            # neowise_log.json tracks which sources we have local data for.  to-do: update log for sufficiently old entries
            data_log = self.read_file(self.save_location + "/neowise_log.json").json()
        except Exception as e:
            self.logger.warn(str(e))
            self.logger.warn(
                "New or no data log (check exception above), downloading all sources from ipac"
            )
            data_log = {}
            buf = io.BytesIO()
            buf.write(bytes(json.dumps(data_log, indent=3), "utf-8"))
            buf.seek(0)
            self.put(self.save_location + "/neowise_log.json", buf.read())
            self.logger.info("created new (blank) neowise_log")
        
        return data_log

    def process(
        self, gen: Generator[TransientView, T3Send, None], t3s: Optional[T3Store] = None
    ) -> Union[UBson, UnitResult]:
        """
        downloads for neoWISE data at IPAC
        this function is based on plot_neoWISE.py from the sZTF repo by sjoertvv.

        plot neo-wise photometry to alert dropbox: {name}-neoWISE.pdf
        back up raw neoWISE data for future matches: {name}-neoWISE.txt
        save in json: info_list, wise_class, out_dict: {name}_neoWISE.json

        time axis is normalized to MJD=58119, Jan-01-2018.

        in the future, we could make this more stand-alone to allow usage by other Ampel teams
        however the need to download the data makes it hard to make this portable.
        """
        selcols_str = "selcols=ra,dec,sigra,sigdec,mjd,w1mpro,w1sigmpro,w1flux,w1sigflux,w1sky,w2flux,w2sigflux,w2mpro,w2sigmpro,w2sky,cc_flags,nb,saa_sep,qual_frame,w1mpro_allwise,w1sigmpro_allwise,w2mpro_allwise,w2sigmpro_allwise,w3mpro_allwise,w3sigmpro_allwise"

        url_to_fill = (
            "https://irsa.ipac.caltech.edu/cgi-bin/Gator/nph-query?catalog=neowiser_p1bs_psd&objstr={0}&radius=2.5&spatial=Cone&radunits=arcsec&"
            + selcols_str
        )
        t0 = 58119
            
        data_log: dict[str, Any] = {}

        transients_total = 0
        transients_in_log = 0
        for transients_total, tran_view in enumerate(gen, 1):
            assert tran_view.stock
            tran_name = tran_view.stock["name"][0]
            assert isinstance(tran_name, str)
            tran_year = "20" + tran_name[3:5]

            year_path = self.save_location + f"/{tran_year}"
            if tran_year not in self.get_files(self.save_location):
                self.create_folder(year_path)
            if tran_name not in self.get_files(year_path):
                self.create_folder(year_path + f"/{tran_name}")

            need_data = True
            if not data_log:
                data_log = self.get_data_log()

            assert (lcs := tran_view.get_lightcurves()) is not None
            ra, dec = lcs[-1].get_values(
                "ra"
            ), lcs[-1].get_values("dec")
            filebase = self.save_location + f"/{tran_year}/{tran_name}/{tran_name}"
            out_dict: Optional[dict] = {}

            needs_plot = not self.exists(
                f"{self.save_location}/{tran_year}/{tran_name}/{filebase}-neowise.pdf"
            )

            if tran_name in data_log:
                transients_in_log += 1

            if (tran_name in data_log) and (needs_plot):
                need_data = False
                self.logger.info(f"reading: {tran_name}_neoWISE.txt")
                try:
                    astro_tab = astropy.io.ascii.read(
                        self.read_file(filebase + "_neoWISE.txt").text
                    )
                except Exception as e:  # This occurs for anachronisms, usually due to testing.
                    self.logger.warn(str(e))
                    self.logger.warn(
                        f"Could not find neoWISE log for source {tran_name}, this should not happen.  Check to make sure the path is correct. Re-downloading..."
                    )
                    need_data = True
            if need_data:
                assert ra is not None
                assert dec is not None
                obs_str = "{0:0.6f}+{1:+0.6f}".format(np.median(ra), np.median(dec))

                url = url_to_fill.format(obs_str)
                if self.verbose:
                    self.logger.info(f"getting neoWISE data from: \n{url}")
                try:
                    r = requests.get(url, timeout=120)

                    content = str(r.content)

                    urlindex = content.find("/workspace/")
                    urlindex2 = content.find(".tbl")

                    table = requests.get(
                        "https://irsa.ipac.caltech.edu"
                        + content[urlindex:urlindex2]
                        + ".tbl"
                    )

                except Exception as e:
                    self.logger.warn(
                        "plot_neoWISE: no connection to IRSA; url was:\n" + url
                    )
                    self.logger.warn(str(e))
                    data_log[tran_name] = False
                    info_list = []
                    wise_class = None
                    out_dict = None

                    continue

                astro_tab = astropy.io.ascii.read(table.text)

                # save the data
                save_fname = f"{filebase}_neoWISE.txt"
                # buf = io.BytesIO()
                # buf.write(bytes(json.dumps(table.text, indent = 3), 'utf-8'))
                # buf.seek(0)
                self.put(save_fname, bytes(table.text, "utf-8"))
                data_log[tran_name] = True

            dd = np.array(astro_tab)

            if len(dd) == 0:
                info_list, wise_class, out_dict = ["neoWISE: no match"], None, None

            # apply some quality cuts
            iqual = (
                (dd["saa_sep"] > 5)
                * (dd["qual_frame"] > 0)
                * (dd["cc_flags"] == "0000")
            )
            # if self.verbose:
            # print ('# neoWISE observations', len(dd))
            # print ('# data point to be removed by quality cuts', sum(iqual==False))

            # print ('apply_qcuts=', self.apply_qcuts)

            if sum(iqual) < 3:
                if self.verbose & (sum(iqual) > 0):
                    self.logger.info(
                        "not enough good data."
                    )  # + np.array(astro_tab[('qual_frame', 'saa_sep', 'cc_flags','w1mpro', 'w1sigmpro')]))
                elif self.verbose:
                    self.logger.info("no data for this source.")
                if self.apply_qcuts:
                    info_list, wise_class, out_dict = (
                        [
                            "neoWISE: not enough detections (N={0})".format(
                                sum(iqual)
                            )
                        ],
                        None,
                        None,
                    )
                    continue

            if self.apply_qcuts:
                dd = dd[iqual]
            else:
                self.logger.info("not applying quality cuts")
            dd = dd[np.argsort(dd["mjd"])]

            xx = dd["mjd"] - t0

            # attempt to make smart bins that are catch the entire light curve
            roundtime = np.unique(np.round(xx, decimals=-2))  # bins in 100 days
            neartime = [
                xx[np.argmin(np.abs(xx - rt))] for rt in roundtime
            ]  # catch nearest data
            bins = []  # find the edges of the bins
            for nt in neartime:
                ii = abs(xx - nt) < 30
                if sum(ii):
                    nb = [min(xx[ii]) - 0.1, max(xx[ii]) + 0.1]  # some padding
                    # avoid duplicates
                    if not nb in bins:
                        bins.append(nb)

            bins = list(np.array(bins).flatten())

            # compute: W1
            ii1 = dd["w1sigmpro"] > 0
            xbin1, ybin1 = self.binthem(
                xx[ii1],
                dd["w1mpro"][ii1],
                dd["w1sigmpro"][ii1],
                use_wmean=True,
                std=True,
                sqrtN=True,
                bins=bins,
                silent=True,
            )
            inz1 = ybin1[0, :] > 0

            if sum(ii1):
                w1_med = np.median(dd["w1mpro"][ii1])
                w1_N = sum(ybin1[3, :])
                # w1_chi2 = sum((dd['w1mpro'][ii1]-w1_med)**2 / dd['w1sigmpro'][ii1]**2) / sum(ii1)  # use all data with reported uncertainty
                inz1_chi = (
                    ybin1[3, :] > 2
                )  # need enough detection to compute scatter
                w1_chi2 = sum(
                    (ybin1[0, inz1_chi] - w1_med) ** 2 / ybin1[1, inz1_chi] ** 2
                ) / sum(
                    inz1_chi
                )  # use uncertainty estimated from scatter in each bin

                if out_dict is not None:
                    out_dict["w1_mag_mean"] = ybin1[0, inz1]
                    out_dict["w1_mag_med"] = w1_med
                    out_dict["w1_mag_rms"] = ybin1[1, inz1]
                    out_dict["w1_N_obs"] = ybin1[3, inz1]
                    out_dict["w1_time"] = xbin1[inz1]

            else:
                w1_N, w1_med, w1_chi2 = 0, np.nan, np.nan

            # plot and compute: W2

            ii2_mag = dd["w2sigmpro"] > 0  # significant detections
            ii2 = dd["w2sigflux"] > 0  # all detections
            if sum(ii2) and sum(ii2_mag):

                zp2 = np.median(
                    dd["w2mpro"][ii2_mag] + 2.5 * np.log10(dd["w2flux"][ii2_mag])
                )

                xbin2, ybin2 = self.binthem(
                    xx[ii2],
                    dd["w2flux"][ii2],
                    dd["w2sigflux"][ii2],
                    use_wmean=True,
                    std=True,
                    sqrtN=True,
                    bins=bins,
                    silent=True,
                )

                inz2 = (
                    ybin2[0, :] > 0
                )  # note, this removed epoch where the mean flux is negative

                # convert mean flux to mag
                w2binflux = ybin2[0, :].copy()
                ybin2[0, :] = -2.5 * np.log10(w2binflux) + zp2
                ybin2[1, :] = 2.5 / np.log(10) * ybin2[1, :] / w2binflux

                w2_med = (
                    -2.5
                    * np.log10(self.wmean(dd["w2flux"][ii2], dd["w2sigflux"][ii2]))
                    + zp2
                )  # convert to mag
                w2_N = sum(ybin2[3, :])

                # w2_chi2 = sum((dd['w2mpro'][ii2]-w2_med)**2 / dd['w2sigmpro'][ii2]**2) / sum(ii2) # use all data with reported uncertainty
                inz2_chi = (
                    ybin2[3, :] > 2
                )  # need enough detection to compute scatter
                w2_chi2 = sum(
                    (ybin2[0, inz2_chi] - w2_med) ** 2 / ybin2[1, inz2_chi] ** 2
                ) / sum(
                    inz2_chi
                )  # use uncertainty estimated from scatter in each bin

                if out_dict is not None:
                    out_dict["w2_mag_mean"] = ybin2[0, inz2]
                    out_dict["w2_mag_med"] = w2_med
                    out_dict["w2_mag_rms"] = ybin2[1, inz2]
                    out_dict["w2_N_obs"] = ybin2[3, inz2]
                    out_dict["w2_time"] = xbin2[inz2]

            else:
                w2_N, w2_med, w2_chi2 = 0, np.nan, np.nan

            # compute W1-W2 in each bin
            w1minw2 = []
            if (sum(ii1) > 10) and (sum(ii2) > 10) and (sum(ii2_mag) > 5):
                for x in xbin2[inz2]:
                    # check that we have data in the same bins
                    i1_near = abs(xbin1[inz1] - x) < 20
                    i2_near = abs(xbin2[inz2] - x) < 20
                    n1_near = sum(ybin1[3, :][inz1][i1_near])
                    n2_near = sum(ybin2[3, :][inz2][i2_near])
                    if (n1_near > 2) and (n2_near > 2):
                        w1_near = np.mean(ybin1[0, :][inz1][i1_near])
                        w2_near = np.mean(ybin2[0, :][inz2][i2_near])
                        w1minw2.append(w1_near - w2_near)

            # do classification
            wise_class = None

            # and make a nice string that we can return
            if w2_N == 0 and w1_N == 0:
                wise_info = "neoWISE: no match"
            else:
                wise_info = "neoWISE: N(w1)={0:0.0f}".format(w1_N)

            # keep the string small
            # if w1_N>0:
            #   wise_info += '; <w1>={0:0.2f}'.format(w1_med)
            # if w2_N>0:
            #   wise_info += '; <w2>={0:0.2f}'.format(w2_med)

            if out_dict is not None:

                if not np.isnan(np.median(w1minw2)):
                    wise_info += "; <w1-w2>={0:0.2f}".format(np.median(w1minw2))
                    out_dict["<w1-w2>"] = np.median(w1minw2)

                if not np.isnan(w1_chi2):
                    wise_info += "; chi2(w1)={0:0.1f}".format(w1_chi2)
                    out_dict["chi2(w1)"] = w1_chi2

                if not np.isnan(w2_chi2):
                    wise_info += "; chi2(w2)={0:0.1f}".format(w2_chi2)
                    out_dict["chi2(w2)"] = w2_chi2

            # we want at least 10 W2 detections over 3 epochs
            if w2_N > 10:
                if sum(ybin2[3, inz2]) > 10 and (len(xbin2[inz2]) > 3):

                    # Stern12
                    # if (w2_med<15) and (np.median(w1minw2)>0.8):
                    #   wise_class = 'AGN'

                    # Assaf+13
                    if np.median(w1minw2) > 0.662 * np.exp(
                        0.232 * (np.clip(w2_med - 13.97, 0, 1e99)) ** 2
                    ):
                        wise_class = "AGN"
                        # wise_info+='; AGN classification based on color (Assef+12)'

                # significant variability
                if w1_chi2 > 10:
                    wise_class = "AGN?"
                    # wise_info+='; significant variability'
                # evidence for variability
                elif w1_chi2 > 5:
                    wise_class = None
                    # wise_info+='; some evidence for variability'

            info_list = [wise_info]

            # also add allWISE info
            if max(dd["w2mpro_allwise"]) and max(dd["w1mpro_allwise"]):
                w1w2_allwise = max(dd["w1mpro_allwise"]) - max(dd["w2mpro_allwise"])
                w1w2_allwise_err = np.sqrt(
                    max(dd["w1sigmpro_allwise"]) ** 2
                    + max(dd["w2sigmpro_allwise"]) ** 2
                )
                all_info = "allWISE: w1-w2={0:0.2f}+/-{1:0.2f}".format(
                    w1w2_allwise, w1w2_allwise_err
                )

                if np.median(w1w2_allwise) > 0.662 * np.exp(
                    0.232 * (np.clip(dd[0]["w2mpro_allwise"] - 13.97, 0, 1e99)) ** 2
                ):
                    wise_class = "AGN"
                    all_info += "; AGN classification based on color (Assef+12)"

                info_list += [all_info]

            save_fname = filebase + "_neoWISE.json"
            buf = io.BytesIO()
            buf.write(
                bytes(
                    json.dumps(
                        self.jsonify(
                            {
                                "info_list": info_list,
                                "wise_class": wise_class,
                                "out_dict": out_dict,
                            }
                        ),
                        indent=3,
                    ),
                    "utf-8",
                )
            )
            buf.seek(0)
            self.put(save_fname, buf.read())

            if self.verbose:
                # print ('allWISE: median(w1)-median(w2) = {0:0.3f}'.format(w1_med-w2_med))
                # [print(x) for x in info_list]
                self.logger.info("neoWISE classification: " + str(wise_class))

            # make a plot
            plt.close()
            fig = plt.figure()
            ax = fig.add_subplot(111)

            if w1_N:
                line = ax.errorbar(
                    xx[ii1],
                    dd["w1mpro"][ii1],
                    dd["w1sigmpro"][ii1],
                    fmt="s",
                    alpha=0.5,
                    label="W1",
                )
                ax.plot(
                    xbin1[inz1], ybin1[0, inz1], "d--", color=line[0].get_color()
                )
                ax.fill_between(
                    xbin1[inz1],
                    ybin1[0, inz1] + ybin1[1, inz1],
                    ybin1[0, inz1] - ybin1[1, inz1],
                    color=line[0].get_color(),
                    alpha=0.5,
                )
                if self.plot_allWISE and max(dd["w1mpro_allwise"]):
                    ax.plot(
                        xx,
                        np.repeat(max(dd["w1mpro_allwise"]), len(xx)),
                        ":",
                        lw=2,
                        color=line[0].get_color(),
                        alpha=0.6,
                    )

            if w2_N:
                line = ax.errorbar(
                    xx[ii2],
                    dd["w2mpro"][ii2],
                    dd["w2sigmpro"][ii2],
                    fmt="o",
                    alpha=0.5,
                    label="W2",
                )
                ax.plot(
                    xbin2[inz2], ybin2[0, inz2], "d--", color=line[0].get_color()
                )
                ax.fill_between(
                    xbin2[inz2],
                    ybin2[0, inz2] + ybin2[1, inz2],
                    ybin2[0, inz2] - ybin2[1, inz2],
                    color=line[0].get_color(),
                    alpha=0.5,
                )
                if self.plot_allWISE and max(dd["w2mpro_allwise"]):
                    ax.plot(
                        xx,
                        np.repeat(max(dd["w2mpro_allwise"]), len(xx)),
                        ":",
                        lw=2,
                        color=line[0].get_color(),
                        alpha=0.6,
                    )

            ax.set_ylim(ax.get_ylim()[::-1])
            ax.set_xlabel("MJD - {0:0.0f}".format(t0))
            ax.set_ylabel("mag (Vega)")
            titstr = "N(w1)={0:0.0f}; N(w2)={1:0.0f}; chi2(w1)={2:0.1f}; chi2(w2)={3:0.1f} <w1-w2>={4:0.2f}".format(
                w1_N, w2_N, w1_chi2, w2_chi2, np.median(w1minw2)
            )
            ax.set_title(titstr, fontsize=11)
            ax.legend()
            fig.tight_layout()

            plot_fname = f"{filebase}-neoWISE.pdf"
            buf = io.BytesIO()
            ax.figure.savefig(buf, format="pdf")
            buf.seek(0)
            self.put(plot_fname, buf.read())
            self.logger.debug(f"plotting {tran_name}")
            ax.clear()

            self.maybe_commit()

        if transients_total > 0:
            self.logger.info(
                f"found {transients_in_log} out of {transients_total} sources in data log"
            )

            buf = io.BytesIO()
            buf.write(bytes(json.dumps(data_log, indent=3), "utf-8"))
            buf.seek(0)
            self.put(self.save_location + "/neowise_log.json", buf.read())

            self.commit()
        return None

    def binthem(
        self,
        x,
        y,
        yerr=None,
        bins=10,
        range=[],
        use_mean=False,
        use_sum=False,
        use_wmean=False,
        cl=0.9,
        poisson=False,
        std=False,
        sqrtN=True,
        smart_expand=False,
        silent=False,
    ):
        """
        >> xmid, ymid = binthem(x, y)
        bin parameter y using bins in x-direction
        output:
         - xmid  (N) the mean value of x for each bin
         - ymid  (4,N) containing the [median/mean/total value, -/+ of uncertainty/dispersion, number in bin]
        input:
         - x,y  equal length arrays
         - bins  number of xbins or array with bins (default is 10)
        optional input:
         - range=[xmin, xmax] range for the bins (default is [min(x),max(x)])
         - cl=0.9  confidence level for computing the uncertainty
         - poisson=False use Poisson stat to find uncertainty on number in bin
         - std=False  use standard deviation / sqrt(N) to compute uncertainty (ie, symtric errorbars)
         - sqrtN=True  set to False to use std only
         - use_mean=False  use mean (median is default)
         - use_sum=False sum bin content
         - use_wmean=False  compute weighted mean, requires yerr= input
         - smart_expand=False, if set, we try to increase bins to catch clusters of data (warning: experimentenal!)
         - silent  shutup
        """

        if use_wmean:
            if yerr is None:
                self.logger.warn("ERROR, please give yerr= input when wmean=True")
                return

        if np.isscalar(bins):
            if len(range) == 2:
                x_bins = np.linspace(range[0], range[1], bins)
            else:
                x_bins = np.linspace(np.min(x), np.max(x), bins)
        else:
            x_bins = bins.copy()

        if smart_expand:

            for i in np.arange(1, len(x_bins)):
                dbin = x_bins[i] - x_bins[i - 1]
                iright = x > x_bins[i]
                if sum(iright):
                    # points closer to the current bins than next bins
                    if min(x[iright] - x_bins[i]) < dbin / 2.0:
                        x_bins[i] = x_bins[i] + dbin / 2.0
        xmid = np.zeros(len(x_bins) - 1)
        ymid = np.zeros((5, len(x_bins) - 1))

        for i in np.arange(len(xmid)):

            ibin = np.where((x >= x_bins[i]) & (x < x_bins[i + 1]))[0]
            if i == (len(xmid) - 1):  # close last bin
                ibin = np.where((x >= x_bins[i]) & (x <= x_bins[i + 1]))[0]

            if len(ibin) > 0:

                xmid[i] = np.mean(x[ibin])
                ymid[4, i] = np.std(x[ibin])

                y_arr = y[ibin]
                ymid[3, i] = len(ibin)

                if use_mean:
                    ymid[0, i] = np.mean(y_arr)
                elif use_sum:
                    ymid[0, i] = np.sum(y_arr)
                elif use_wmean:
                    y_arr_err = yerr[ibin]
                    ymid[0, i] = self.wmean(y_arr, 1 / y_arr_err * 2)
                else:
                    ymid[0, i] = np.median(y_arr)

                if std:
                    ymid[[1, 2], i] = np.std(y_arr)
                    if sqrtN:
                        ymid[[1, 2], i] = ymid[[1, 2], i] / np.sqrt(len(ibin))
                elif poisson and (len(ibin) > 1):
                    ymid[[1, 2], i] = np.abs(
                        poisson_limits(np.sum(y[ibin]), cl) / len(ibin) - ymid[0, i]
                    )
                elif use_wmean:
                    ymid[[1, 2], i] = 1 / np.sqrt(sum(1 / y_arr_err**2))
                else:
                    y_arrs = np.sort(y_arr)
                    ii = np.arange(len(ibin))
                    ymid[1, i] = np.abs(
                        ymid[0, i] - np.interp((cl / 2.0 - 0.5) * len(ibin), ii, y_arrs)
                    )
                    ymid[2, i] = np.abs(
                        ymid[0, i] - np.interp((cl / 2.0 + 0.5) * len(ibin), ii, y_arrs)
                    )
            else:
                xmid[i] = (x_bins[i] + x_bins[i + 1]) / 2.0

        if sum(ymid[3, :]) > len(x):
            self.logger.warn(
                "binthem: WARNING: more points in bins ({0}) compared to lenght of input ({1}), please check your bins".format(
                    sum(ymid[3, :]), len(x)
                )
            )
            # key = input()

        return xmid, ymid

    def wmean(self, d, ivar):
        """
        >> mean = wmean(d, ivar)
        inverse-variance weighted mean
        """
        d = np.array(d)
        ivar = np.array(ivar)
        return np.sum(d * ivar) / np.sum(ivar)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/nuclear/t3/T3MetricsPlots.py
# License           : BSD-3-Clause
# Author            : mitchell@nyu.edu
# Date              : 08.06.2020
# Last Modified Date: 29.11.2022
# Last Modified By  : simeon.reusch@desy.de

import io, time, json

from typing import Tuple, Dict, List, Any, Optional, Union
from collections.abc import Generator

import numpy as np
from pydantic import BaseModel
import astropy  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
import tempfile

from ampel.types import UBson, T3Send
from ampel.view.TransientView import TransientView
from ampel.struct.T3Store import T3Store
from ampel.view.LightCurve import LightCurve
from ampel.struct.UnitResult import UnitResult
from ampel.abstract.AbsPhotoT3Unit import AbsPhotoT3Unit
from ampel.nuclear.flexfit import flexfit
from ampel.nuclear.t3.classifyme import jsonify
from ampel.nuclear.t2.T2FlexFit import get_raw_lc
from ampel.nuclear.t3.dropboxIO import DropboxUnit

dump_keys = [
    "name",
    "ampelID",
    "dropbox_path",  # :"alerts/2019/"" or False if no path
    "ra",
    "dec",
    "magpsf_r",
    "magpsf_g",
    "magnr_r",
    "magnr_g",
    "flexclass",
    "neoWISE_chi2_w1",
    "neoWISE_chi2_w2",
    "allWISE_w1-w2",
    "neoWISE_w1-w2",
    "catalogs_matched",  # ["PS1", "Gaia", ...]
    "Gaia_compactness",
]


def get_t2_result_with_compound(
    view: TransientView, unit_id: str
) -> Tuple[Optional[Dict[str, Any]], Optional[Union[int, bytes, str]]]:

    t2_res = reversed(list(view.get_t2_views(unit=unit_id) or []))

    for t2 in t2_res:
        result = t2.get_payload()

        if result and t2.code >= 0:
            assert isinstance(result, dict)
            return result, t2.link

    return None, None


def get_lightcurve(
    view: TransientView, compound_id: Union[int, bytes, str]
) -> Optional[LightCurve]:
    for lc in reversed(view.get_lightcurves() or []):
        if lc.compound_id == compound_id:
            return lc
    return None


class T3MetricsPlots(DropboxUnit):

    verbose: bool = False  #: give feedback on what is plotted

    def post_init(self):
        """
        Run config keys: save_location, verbose, mode
        """
        # super().post_init()
        self.save_location = self.base_location + "/alerts"
        if "alerts" not in self.get_files(self.base_location):
            self.create_folder(self.save_location)
        self.dump_log = []

    def process(
        self, gen: Generator[TransientView, T3Send, None], t3s: Optional[T3Store] = None
    ) -> Union[UBson, UnitResult]:
        """
        Loops through transients and plots results of fit, offset plots, and dumps fit info, simple metrics, and raw lightcurve
        """

        for tran_view in gen:
            assert tran_view.stock is not None
            tran_name = tran_view.stock["name"][0]
            assert isinstance(tran_name, str)
            tran_year = '20'+tran_name[3:5]
            year_path = self.save_location + f"/{tran_year}"
            # NB: this is cached
            if tran_year not in self.get_files(self.save_location):
                self.create_folder(year_path)
            entries_for_year = self.get_files(year_path)
            self.dump_log.append({})
            self.dump_log[-1]["name"] = tran_name
            
            filepath = year_path + f"/{tran_name}/{tran_name}"
            self.dump_log[-1]["dropbox_path"] = filepath

            if tran_name not in entries_for_year:
                self.create_folder(year_path + f"/{tran_name}")
            
            flex, compound_id = get_t2_result_with_compound(
                        view=tran_view, unit_id="T2FlexFit"
                    )
            simple = [
                dv.get_payload()
                for dv in tran_view.get_t2_views(unit="T2SimpleMetrics")
            ][0]

            ps1_match_dvs = tran_view.get_t2_views(unit="T2CatalogMatch")
            ps1_match = [dv.get_value("PS1", dict) for dv in ps1_match_dvs][0]

            if (
                flex is not None
                and compound_id is not None
                and (lc := get_lightcurve(tran_view, compound_id)) is not None
            ):
                dict_lc = get_raw_lc(lc)
                dump_fname = f"{filepath}_lc.json"
                buf = io.BytesIO()
                buf.write(
                    bytes(json.dumps(jsonify(dict_lc), indent=3), "utf-8")
                )
                buf.seek(0)
                self.put(dump_fname, buf.read())

            if flex and "fit_params" in flex.keys():
                if flex["fit_params"]["photoclass"] != "no fit":
                    self.fit_plot(flex, filepath)
                    if self.verbose:
                        self.logger.info(
                            "successfully plotted flex fit for source {}".format(
                                tran_name
                            )
                        )
                    dump_fname = f"{filepath}_flex.json"
                    buf = io.BytesIO()
                    buf.write(
                        bytes(json.dumps(flex["fit_params"], indent=3), "utf-8")
                    )
                    buf.seek(0)
                    self.put(dump_fname, buf.read())

                else:
                    self.logger.info(
                        "no flex fit for source {}".format(tran_name)
                    )
            else:
                self.logger.warn(
                    "empty flex fit for source {}".format(tran_name)
                )

            if ps1_match and isinstance(simple, dict):
                if "metrics" in simple.keys():
                    (
                        dra_ps1,
                        ddec_ps1,
                        offset_med_ps1,
                        offset_rms_ps1,
                        offset_sig_ps1,
                    ) = self.get_ps1(simple, ps1_match, filepath)
                    simple = dict(simple)
                    simple["plot_info"] = {
                        "dra_ps1": dra_ps1,
                        "ddec_ps1": ddec_ps1,
                        **simple["plot_info"],
                    }
                    simple["metrics"] = {
                        "offset_med_ps1": offset_med_ps1,
                        "offset_rms_ps1": offset_rms_ps1,
                        "offset_sig_ps1": offset_sig_ps1,
                        **simple["metrics"],
                    }

            if isinstance(simple, dict):
                if "metrics" in simple.keys():
                    self.offset_plots(simple, filepath)
                    if self.verbose:
                        self.logger.info(
                            "successfully plotted offset plots for source {}".format(
                                tran_name
                            )
                        )

                    dump_fname = f"{filepath}_simple.json"
                    buf = io.BytesIO()
                    buf.write(
                        bytes(json.dumps(simple["metrics"], indent=3), "utf-8")
                    )
                    buf.seek(0)
                    self.put(dump_fname, buf.read())

                else:
                    self.logger.info(
                        "no simple metrics for source {}".format(tran_name)
                    )
            else:
                self.logger.warn(
                    "empty simple metrics for source {}".format(tran_name)
                )
            
            self.maybe_commit()
        self.commit()
        return None

    def get_ps1(self, simple, ps1_match, filepath):
        metrics, plot_info = simple["metrics"], simple["plot_info"]
        ps1_ra, ps1_dec = ps1_match["RA"], ps1_match["Dec"]
        dra_ps1 = (np.array(metrics["ra"]) - ps1_ra) * 3600  # mean dra
        ddec_ps1 = (np.array(metrics["dec"]) - ps1_dec) * 3600  # mean dec
        dstd_ps1 = np.clip(
            np.std(np.append(dra_ps1, ddec_ps1)) / np.sqrt(len(dra_ps1)), 0.1, 1
        )  # double-check this!
        offset_med_ps1 = np.sqrt(np.median(dra_ps1) ** 2 + np.median(ddec_ps1) ** 2)
        offset_rms_ps1 = np.std(np.append(dra_ps1, ddec_ps1))
        offset_sig_ps1 = offset_med_ps1 / dstd_ps1

        # self.logger.info('simple_metrics: distance to PS1 (median, rms, significance) : {0:0.3f} {1:0.3f} {2:0.3f}'.format(
        #   offset_med_ps1, offset_rms_ps1, offset_sig_ps1))

        return dra_ps1, ddec_ps1, offset_med_ps1, offset_rms_ps1, offset_sig_ps1

    def fit_plot(self, result, filepath):
        fit_params, lc_data, plot_info = (
            result["fit_params"],
            result["lightcurve_data"],
            result["plot_info"],
        )

        plt.close()
        fig = plt.figure()
        ax = fig.add_subplot(111)

        filters = ["g", "r"]
        ax.clear()

        if plot_info:

            dtime_out = plot_info["dtime_out"]
            if (dtime_out != None) and (plot_info["flux_out"] != None):
                ax.plot(dtime_out, plot_info["flux_out"], "sk", zorder=1e99)

            dtime, flux, flux_err, fid, isdetect = (
                np.array(lc_data["dtime"]),
                np.array(lc_data["flux"]),
                np.array(lc_data["flux_err"]),
                np.array(lc_data["fid"]),
                np.array(lc_data["isdetect"]),
            )
            lsq_out = plot_info["lsq_out"]
            dt_fix = plot_info["dt_fix"]
            photoclass = fit_params["photoclass"]
            name = fit_params["name"]
            try:
                classification = fit_params["classification"]
            except KeyError:
                classification = photoclass

            xx = np.linspace(min(dtime), max(dtime), 500)
            ax.set_ylim(0, max(flux[isdetect]) * 1.2)
            ax.set_xlim(min(xx), max(xx))
            ax.set_xlabel("Days since first detection")
            ax.set_ylabel("Flux")

            islim = isdetect == False
            for k in filters:
                try:  # handling missing color fits w/ exceptions, could also do this in T2
                    iflt = fid == k
                    ll = ax.errorbar(
                        dtime[isdetect * iflt],
                        flux[isdetect * iflt],
                        flux_err[isdetect * iflt],
                        fmt="s",
                        color=k,
                    )
                    ax.errorbar(
                        dtime[islim * iflt],
                        flux[islim * iflt],
                        flux_err[islim * iflt],
                        alpha=0.3,
                        fmt="v",
                        color=ll[0].get_color(),
                    )
                    ax.plot(
                        xx,
                        flexfit.broken_gauss(lsq_out[k][0], xx, dt_fix=dt_fix[k]),
                        "--",
                        color=k,
                        lw=1.3,
                        alpha=0.7,
                    )
                    ax.set_ylim(0, max(flux[isdetect] * 1.2))
                except:
                    self.logger.warn("no fit for {} band for source {}".format(k, name))

            xx_0 = np.linspace(min(dtime), max(dtime), sum(fid == filters[0]) * 10)
            xx_1 = np.linspace(min(dtime), max(dtime), sum(fid == filters[1]) * 10)
            xx = np.append(xx_0, xx_1)
            fid_in = np.append(
                np.repeat(filters[0], len(xx_0)), np.repeat(filters[1], len(xx_1))
            )

            # x_peakc arguement needed to make sure we use the time normaliation for the color function
            if "color" in lsq_out.keys():
                model_out = flexfit.broken_gauss_twocolor(
                    lsq_out["color"][0],
                    xx,
                    dt_fix=dt_fix["color"],
                    fid=fid_in,
                    x_peakc=np.median(dtime),
                )
                for flt in filters:
                    ax.plot(
                        xx[fid_in == flt],
                        model_out[fid_in == flt],
                        "-",
                        color=flt,
                        lw=2.0,
                        alpha=0.7,
                    )
            else:
                self.logger.warn("no color calculated for source {}".format(name))

            txt = ax.text(
                0.03,
                0.92,
                plot_info["finfo"],
                transform=ax.transAxes,
                verticalalignment="center",
                fontsize=11,
            )
            txt.set_bbox(dict(facecolor="white", alpha=0.5, edgecolor="k"))

            ax.set_xlim(-8, max(np.array(lc_data["jd"]) - np.array(lc_data["jd0"])) + 2)
            ax.set_title(
                fit_params["name"]
                + " ("
                + str(classification).strip()
                + " - "
                + photoclass
                + ")"
            )
        else:
            txt = ax.text(
                0.03,
                0.92,
                "Fewer than 3 observations for source",
                transform=ax.transAxes,
                verticalalignment="center",
            )
            txt.set_bbox(dict(facecolor="white", alpha=0.5, edgecolor="k"))

            try:
                classification = fit_params["classification"]
            except KeyError:
                classification = fit_params["photoclass"]

            ax.set_title(
                fit_params["name"]
                + " (no classification - "
                + fit_params["photoclass"]
                + ")"
            )

        plot_fname = f"{filepath}_flex.pdf"
        buf = io.BytesIO()
        ax.figure.savefig(buf, format="pdf")
        buf.seek(0)
        self.put(plot_fname, buf.read())
        ax.clear()

    def offset_plots(self, simple, filepath):
        plt.close()
        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_subplot(111)

        metrics, plot_info = simple["metrics"], simple["plot_info"]

        if plot_info["offset_plot"]:
            dra, ddec, snr_weight = (
                np.array(plot_info["dra"]),
                np.array(plot_info["ddec"]),
                np.array(plot_info["snr_weight"]),
            )
            ax.plot(
                dra, ddec, "s", markerfacecolor="none", markeredgecolor="k", label=""
            )

            for band in "gr":
                if f"offset_weighted_{band}" in metrics:
                    mask = np.asarray(plot_info[f"i{band}"])
                    weighted_dra = self.wmean(dra[mask], snr_weight[mask])
                    weighted_ddec = self.wmean(ddec[mask], snr_weight[mask])
                    ax.plot(
                        dra[mask],
                        ddec[mask],
                        "s",
                        markerfacecolor=band,
                        markeredgecolor="k",
                        label="IPAC {0} (mean={1:0.3f}, rms={2:0.3f}, sig={3:0.2f})".format(
                            band,
                            metrics[f"offset_weighted_{band}"],
                            metrics[f"offset_rms_{band}"],
                            metrics[f"offset_sig_{band}"],
                        ),
                    )
                    ax.plot(
                        weighted_dra, weighted_ddec, f"*{band}", markeredgecolor="k"
                    )

            if "offset_med_ps1" in metrics:
                dra_ps1, ddec_ps1 = np.array(plot_info["dra_ps1"]), np.array(
                    plot_info["ddec_ps1"]
                )
                ax.plot(
                    dra_ps1,
                    ddec_ps1,
                    "o",
                    markerfacecolor="orange",
                    markeredgecolor="k",
                    label="PS1     (med ={0:0.3f}, rms={1:0.3f}, sig={2:0.2f})".format(
                        metrics["offset_med_ps1"],
                        metrics["offset_rms_ps1"],
                        metrics["offset_sig_ps1"],
                    ),
                )
                ax.plot(
                    np.median(dra_ps1),
                    np.median(ddec_ps1),
                    "*",
                    markerfacecolor="orange",
                    markeredgecolor="k",
                )

            ax.set_xlabel("RA - <RA>")
            ax.set_ylabel("Dec - <Dec>")
            ax.set_xlim(-1.3, 1.3)
            ax.set_ylim(-1.3, 1.3)
            ax.legend(loc=0)

            name = metrics["name"]
            # ax.set_title(name+' ('+str(self.classification).strip()+')') #to-do: when I have marshal
            ax.set_title(name)

            plot_fname = f"{filepath}_offset.pdf"
            buf = io.BytesIO()
            ax.figure.savefig(buf, format="pdf")
            buf.seek(0)
            self.put(plot_fname, buf.read())
            ax.clear()

    def wmean(self, d, ivar):
        """
        >> mean = wmean(d, ivar)
        inverse-variance weighted mean
        """
        d = np.array(d)
        ivar = np.array(ivar)
        return np.sum(d * ivar) / np.sum(ivar)

    def format_date(self, transient_date):
        return astropy.time.Time(
            transient_date[6:10]
            + "-"
            + transient_date[3:5]
            + "-"
            + transient_date[0:2]
            + transient_date[10:]
        ).jd



"""
function to compute simple light curve metrics
"""
import os
import numpy as np
import pandas as pd  # type: ignore
import json
import astropy
import warnings

from . import flexfit


name = "flex"

dt = [
    ("name", "U35"),
    ("classification", "U35"),
    ("photoclass", "U35"),
    ("blabber", "U200"),
    ("band", "U6"),
    ("rms", "f8"),
    ("mad", "f8"),
    ("chi2", "f8"),
    ("mjd_peak", "f8"),
    ("dtime_peak", "f8"),
    ("e_dtime_peak", "f8"),
    ("mag_peak", "f8"),
    ("e_mag_peak", "f8"),
    ("sigma_rise", "f8"),
    ("sigma_fade", "f8"),
    ("e_sigma_rise", "f8"),
    ("e_sigma_fade", "f8"),
    ("n_rise", "f8"),
    ("n_fade", "f8"),
    ("mean_color", "f8"),
    ("color_slope", "f8"),
    ("e_mean_color", "f8"),
    ("e_color_slope", "f8"),
]


# magnitude/flux conversion with some zp
zp = 26


def flux2mag(flux, zp=zp):
    return -2.5 * np.log10(flux) + zp


def mag2flux(mag, zp=26):
    return 10 ** (-0.4 * (mag - zp))


def jdtomjd(jd):
    return jd - 2400000.5


exception_dict = {}
exception_dict["ZTF17aaazdba"] = {
    "jd_min": 2458515
}  # has early bogus data, so we require jd larger than jd_min


def compute(meta_data, lc):
    """
    Returns {'fit_params':
    ['dtime_peak', 'band', 'mag_peak', 'sigma_rise', 'sigma_fade', 'mean_color', 'color_slope', 'e_dtime_peak',
    'e_mag_peak', 'e_mean_color', 'e_color_slope', 'e_sigma_rise', 'mad', 'rms', 'chi2', 'n_rise', 'n_fade',
    'classification', 'name', 'mjd_peak', 'photoclass', 'blabber'],

    'lc_data':
    ['dtime', 'flux', 'flux_err', 'fid', 'isdetect', 'jd', 'jd0', 'pid', 'programid'],

    'plot_info':
    ['lsq_out', 'dt_fix', 'dtime_out', 'flux_out', , 'pid', 'programid', 'finfo']
    }
    """

    # allow some manual trimming
    if meta_data["name"] in exception_dict:
        edict = exception_dict[meta_data["name"]]
    else:
        edict = None

    lc_data, fit_params, plot_info = {}, {}, {}
    logger = meta_data["logger"]
    dtime, flux, flux_err, fid, isdetect, jd0, magnr_dict, mdata = trim_ztf_ampel_data1(
        lc,
        logger,
        exception_dict=edict,
        oldest_upper_limits=meta_data["oldest_upper_limits"],
        max_post_peak=meta_data["max_post_peak"],
    )
    (
        lc_data["dtime"],
        lc_data["flux"],
        lc_data["flux_err"],
        lc_data["fid"],
        lc_data["isdetect"],
    ) = (dtime, flux, flux_err, fid, isdetect)

    if sum(isdetect) < 3:
        lc_data["jd"] = mdata["jd"]
        lc_data["jd0"] = jd0
        lc_data["pid"] = mdata["pid"]
        lc_data["programid"] = mdata["programid"]

        fit_params = {}
        fit_params["name"] = meta_data["name"]
        fit_params["photoclass"] = "not enough data"
        return {
            "fit_params": fit_params,
            "lightcurve_data": lc_data,
            "plot_info": None
            # {'lsq_out' : None,
            # 'dt_fix': None,
            # 'dtime_out': None,
            # 'flux_out': None,
            # 'pid': None,
            # 'programid': None,
            # 'finfo': None}
        }

    # run flexfit, get dictionary with fit info
    fit_params, lc_data, plot_info = flexfit.flex_fit_wclip(
        dtime, flux, flux_err, fid, logger, isdetect, verbose=False
    )
    lc_data["jd"] = mdata["jd"]
    lc_data["jd0"] = jd0
    lc_data["pid"] = mdata["pid"]
    lc_data["programid"] = mdata["programid"]

    # add some meta data
    fit_params["name"] = meta_data["name"]

    # collect some info to show on plots
    if not "e_dtime_peak" in fit_params:
        str_em = "fix"
    elif fit_params["e_dtime_peak"] < 0:
        str_em = "fix"
    else:
        str_em = "{0:0.1f}".format(fit_params["e_dtime_peak"])

    if not "dtime_peak" in fit_params:
        fit_params["dtime_peak"] = -999
    if "mag_peak" in fit_params:
        finfo = "t_peak={0:0.1f} ({1}); m_peak={2:0.1f}; rise={3:0.1e}; fade={4:0.1e}; chi2={5:0.1f}".format(
            fit_params["dtime_peak"],
            str_em,
            fit_params["mag_peak"],
            fit_params["sigma_rise"],
            fit_params["sigma_fade"],
            fit_params["chi2"],
        )
    else:
        finfo = "no fit"

    if "mean_color" in fit_params:
        if not "e_color_slope" in fit_params:
            fit_params["e_color_slope"] = -999
        if not "e_mean_color" in fit_params:
            fit_params["e_mean_color"] = -999
        finfo += "\ncolor={0:0.2f} +/- {1:0.2f}; slope={2:0.3f} +/- {3:0.3f}".format(
            fit_params["mean_color"],
            fit_params["e_mean_color"],
            fit_params["color_slope"],
            fit_params["e_color_slope"],
        )

    logger.info(f"flex_metrics: {finfo}")
    plot_info["finfo"] = finfo

    # ---
    # get the class, plus some extra chatter
    blab, photoclass = flexfit.flex_class(fit_params)

    logger.info(f"flex_metrics: {photoclass}")
    logger.info(f"flex_metrics: {blab}")

    if jd0:
        fit_params["mjd_peak"] = jdtomjd(fit_params["dtime_peak"] + jd0)
    fit_params["photoclass"] = photoclass
    fit_params["blabber"] = blab

    return {
        "fit_params": fit_params,
        "lightcurve_data": lc_data,
        "plot_info": plot_info,
    }


def trim_ztf_ampel_data1(
    data_json,
    logger,
    oldest_upper_limits=14,
    filters=[1, 2],
    only_ztf=True,
    instruments=["ZTF", "SEDM", "IOO"],
    max_post_peak=200,
    remove_neg=True,
    exception_dict=exception_dict,
):
    """
    convenience function for dealing with data from the ampel

    >>> dtime, flux, flux_err, fid, isdetect, jd0, magnr_dict0, data = trim_ztf_ampel_data(input_json)

    input: dict created by reading output of uploaded_photometry

    upper limits (5-sigma) are stored in flux_err and have flux=0.
    for these limits I try to account for the flux of host galaxy (magnr), I we assume the event is nuclear

    options:
    - filter=[g,r], but can be other filters ;
    - only_ZTF=True, set False to include data from intruments;
    - instruments=['ZTF','SEDM', 'IOO']
    """

    dt = [
        ("name", "U35"),
        ("classification", "U35"),
        ("photoclass", "U35"),
        ("blabber", "U200"),
        ("band", "U6"),
        ("rms", "f8"),
        ("mad", "f8"),
        ("chi2", "f8"),
        ("mjd_peak", "f8"),
        ("dtime_peak", "f8"),
        ("e_dtime_peak", "f8"),
        ("mag_peak", "f8"),
        ("e_mag_peak", "f8"),
        ("sigma_rise", "f8"),
        ("sigma_fade", "f8"),
        ("e_sigma_rise", "f8"),
        ("e_sigma_fade", "f8"),
        ("n_rise", "f8"),
        ("n_fade", "f8"),
        ("mean_color", "f8"),
        ("color_slope", "f8"),
        ("e_mean_color", "f8"),
        ("e_color_slope", "f8"),
    ]

    data_dict1 = {
        key: value
        for key, value in data_json.items()
        if len(value) == len(data_json["magnr"])
    }  # everything that isn't a single value

    # I expect to see RuntimeWarnings in this block, median([])=nan
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        magnr_dict0 = {
            str(k): np.median(
                data_json["magnr"][
                    np.where(
                        (data_json["fid"] == k) & (data_json["magnr"] is not None)
                    )[0]
                ]
            )
            for k in (1, 2, 3)
        }

    data0 = pd.DataFrame(data_dict1)
    isdetect0 = data0["magpsf"] < 23
    islim0 = isdetect0 == False

    # some trimming
    if remove_neg:
        itrim0 = (data0["isdiffpos"].values == "t") | (
            data0["isdiffpos"].values == "1"
        )  # only positive detections
        # print('neg:', itrim0)
    else:
        itrim0 = np.repeat(True, len(data0))

    # don't use upper limits
    # (we can insert them later, after figure out how to account for host contribution adding to the uncertainty)
    itrim0 &= isdetect0

    # selection of bands
    itrim0 &= np.array([x in filters for x in data0["fid"]])

    # option to look only close to peak
    itrim0 &= (
        data0["jd"] - data0["jd"][data0["magpsf"].idxmin()]
    ) < max_post_peak  # upto xx days post guessed peak

    if exception_dict is not None:
        if "jd_min" in exception_dict:
            itrim0 &= data0["jd"] > exception_dict["jd_min"]

    # apply trimming
    data = data0[itrim0]
    # sort
    data = data.sort_values("jd")

    # get the limits and detections
    fid = np.array(
        [
            "g" if d == 1 else "r" if d == 2 else "i" if d == 3 else "check filter?"
            for d in data["fid"]
        ]
    )
    isdetect = data["magpsf"].values < 22
    islim = isdetect == False

    logger.info(
        "# data point used for fitting {0} (with {1} upper limits)".format(
            len(data), sum(islim)
        )
    )

    flux = mag2flux(data["magpsf"].values)
    flux_lim = mag2flux(
        data["diffmaglim"].values
    )  # 5 sigma limit for entire image (ie, not including host noise)
    flux[islim] = 0
    flux_err = np.log(10) / 2.5 * np.clip(data["sigmapsf"].values, 0.1, 2) * flux

    if remove_neg == False:
        ineg = data["isdiffpos"].values == 0
        flux[ineg] *= -1

    # try to guess the real flux upper limit by including the host contribution
    magnr_arr = np.array([magnr_dict0.get(k) for k in fid])
    magnr_arr[magnr_arr == None] = 25  # we dont have the host magnitude, dont use it...
    gain_corr_est = 5.0  # tuned for ZTF

    # flux_err = gain_corr_est * ( (flux_err)**2 + mag2flux(magnr_arr))**0.5 # not clear if this correction is needed, but I have seen residuals for ZTF difference photometry ontop of bright galaxies
    flux_err[islim] = (
        (flux_lim[islim] / 5) ** 2 + gain_corr_est**2 * mag2flux(magnr_arr[islim])
    ) ** 0.5

    if sum(isdetect):
        jd0 = min(data[isdetect]["jd"])
        dtime = data["jd"].values - jd0  # pick time normalization
    else:
        jd0 = None
        dtime = None

    return dtime, flux, flux_err, fid, isdetect, jd0, magnr_dict0, data.to_records()



from ampel.config.builder.DistConfigBuilder import DistConfigBuilder
from ampel.config.builder.DisplayOptions import DisplayOptions


def test_build_config() -> None:
    cb = DistConfigBuilder(DisplayOptions(verbose=True, debug=True))
    cb.load_distributions()
    assert not cb.error
    config = cb.build_config(
        stop_on_errors=2,
        config_validator="ConfigValidator",
        get_unit_env=False,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/nuclear/t3/classifyme.py
# License           : BSD-3-Clause
# Author            : mitchell@nyu.edu & sjoert@nyu.edu
# Date              : 08.06.2020
# Last Modified Date: 11.12.2020
# Last Modified By  : sjoert@nyu.edu

"""
container with a function to get classification information 
this is needed for mulitple T3s of ZTFbh
"""

import json
import io


def jsonify(obj):
    """
    Recursively cast to JSON native types
    """
    if hasattr(obj, "tolist"):
        return obj.tolist()
    elif hasattr(obj, "keys"):
        return {jsonify(k): jsonify(obj[k]) for k in obj.keys()}
    elif hasattr(obj, "__len__") and not isinstance(obj, str):
        return [jsonify(v) for v in obj]
    elif hasattr(obj, "isoformat"):
        return obj.isoformat()
    else:
        return obj


def get_class(tran_view, dropbox, logger, write=False):
    """
    Collect classifications from T2 transient view info,
    and our own neoWISE measurment.

    Input is:
     - trans_view: the t2 output which contains the catalogs matches and Fritz/TNS/Growth info
     - dropbox: instance to read/write (from dropboxIO, inititated by each T2)
     - logger: to keep chatting

    The following order applies to accepting classification:
     Fritz
     Growth
     TNS
     static catalog matches from T2
     neoWISE variability measurments form our T3

    This function also returns some extra information about the neoWISE measurements.

    >> class_str, extra_info = get_class(tran_view dropbox, logger, write=True)

    """

    growth_report = tran_view.extra.get("GROWTHMarshalReport")
    tns_report = reports[0] if (reports := tran_view.extra.get("TNSReports")) else None
    fritz_report = tran_view.extra.get("FritzReport")

    alert_base = (
        "/mampel/alerts/" + "20" + tran_view.stock["name"][0][3:5]
    )  # lets assume ZTF ends before 2099 :)

    # read  WISE light curve and color information
    wise_path = alert_base + "/{0}/{0}_neoWISE.json".format(tran_view.stock["name"][0])
    try:
        infile = dropbox.read_file(wise_path)
        if isinstance(infile, str):
            wise_json = json.loads(infile)
        else:
            wise_json = infile.json()
        logger.info("loading wise_class for {}".format(tran_view.stock["name"][0]))
        wise_class = wise_json["wise_class"]
        wise_info = wise_json["info_list"][
            0
        ]  # this now this just contains the neoWISE light curve info
    except FileNotFoundError:
        wise_class = None
        wise_info = ""

    # read the t2 records, take care of rouge dicts that mess with json
    # ampel_matches = tran_view.get_t2_result("T2CatalogMatch") or {}
    ampel_matches = tran_view.get_latest_t2_body(unit="T2CatalogMatch")

    xmatches = {}
    for key in ampel_matches:
        try:
            xmatches[key] = json.loads(json.dumps(ampel_matches[key]))
        # only report distance if rest of content cant be jsonified
        except TypeError:
            xmatches[key] = {"dist2transient": ampel_matches[key]["dist2transient"]}

    # write classifications
    if write:
        class_match = {
            "GROWTH": growth_report,
            "TNS": tns_report,
            "Fritz": fritz_report,
            "xmatch": xmatches,
        }

        buf = io.BytesIO()
        buf.write(bytes(json.dumps(jsonify(class_match), indent=3), "utf-8"))
        buf.seek(0)
        dropbox.put(
            alert_base + "/" + tran_view.stock["name"][0] + "/class_match.json",
            buf.read(),
        )

    # ---
    # Classify

    # if we have a Fritz classification, it is selected first ...
    if fritz_report:
        if "classifications" in fritz_report:

            # Frits can have mulitple classifications,
            # for now, just pick the first one
            if classification := next(
                iter(fritz_report["classifications"] or []), None
            ):

                # shorten some labels
                fritz_class = classification["classification"]
                fritz_class = fritz_class.replace("Tidal Disruption Event", "TDE")

                # deal with lower probability classifications
                if (prob := classification["probability"]) is not None and prob < 0.9:
                    fritz_class += "?"

                    return fritz_class, wise_info

    # ... otherwise Marshal ...
    if growth_report:
        if growth_report["classification"] is not None:
            return growth_report["classification"], wise_info

    # ... then TNS ...
    if tns_report:
        if tns_report["object_type"]["name"] is not None:
            return tns_report["object_type"]["name"], wise_info

    # ... or finally, do classification based on catalog matches
    classification = "None"
    if ampel_matches["milliquas"]:
        classification = "AGN"

    if ampel_matches["wise_color"]:
        classification = "AGN"

    if ampel_matches["varstars"]:
        if classification == "None":
            classification = "varstar"
        elif classification == "AGN":
            classification = "QSO"

    # ... or finally finally, use my neoWISE class
    if wise_class is not None:
        if classification == "None":
            classification = wise_class

    return classification, wise_info



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/nuclear/t0/NuclearFilter.py
# License           : BSD-3-Clause
# Author            : sjoertvv <sjoert@umd.edu>
# Date              : 26.02.2018
# Last Modified Date: 29.11.2022
# Last Modified By  : sr <simeon.reusch@desy.de>

from typing import Optional, Union, List, Dict, Any

import numpy as np
import astropy.coordinates  # need for l, b computation

from ampel.protocol.AmpelAlertProtocol import AmpelAlertProtocol
from ampel.abstract.AbsAlertFilter import AbsAlertFilter

from ampel.nuclear.t0.GaiaVeto import GaiaVetoMixin

REJECTION_REASON_CODES = {
    999: "Passed",
    10: "Large sgscore",
    11: "Large sgscore2",
    12: "No sgscore available",
    20: "Bad PS1 photometry",
    21: "PS1 too far",
    22: "Bright 1st PS1 match",
    23: "Bright 2nd PS1 match",
    24: "Bright 3rd PS1 match",
    30: "Nothing passed default filter",
    40: "Not enough detections",
    50: "Too bright in ZTF reference image",
    60: "Too many PS1 matches (field too crowded)",
    61: "Too close to Galactic Plane (field too crowded)",
    70: "RB too low",
    80: "Only duplicate detections",
    90: "Potential mover",
    100: "Large gap between detections",
    110: "Too faint",
    120: "Flux increase too small",
    130: "Gaia veto",
    140: "Distance veto",
    141: "No distance could be computed",
    150: "Last detection did not pass default filter (only when lastOnly=True)",
}


class NuclearFilter(AbsAlertFilter, GaiaVetoMixin):
    """
    Filter for a clean sample of nuclear transients in galaxies
    includes a veto of PS1 star-galaxy score using Gaia data
    """

    maxDeltaRad: float = 0.5  #: the max flare-ref distance to be accepted, we try mulitple ways of summing distnr for multiple detection
    minDeltaRad: float = 0.0  #: the min flare-ref distance to be accepted

    maxSgscore: float = 0.3  #: min star-galaxy score

    minDeltaJD: float = 0.01  #: remove movers with min time distance in days between detections that pass default filter

    maxDeltaJD: float = 10.0  #: remove lightcurves with gaps, we require we have at least two detections seperated by this distance

    minDetections: int = 3  #: number of detections in each band

    diffmagLimit: float = 20  #: magnitude limit applied to difference flux (magpsf)

    maxDeltaMag: float = 2.5  #: lower limit to flux increase: PSF magnitude in difference image minus PSF magnitude in the references

    minRealBogusScore: float = 0.3  #: min RealBogus score of *any* observation

    maxDistPS1source: float = 1  #: min distance to nearest PS1 source (useful for removing SNe with detections in the ref image, bright stars, and ghostly things)

    closePS1dist: float = 0.5  #: distance when we consider the second PS1 match into our checks for the sgscore and photometry (0.5" recommended)

    brightRefMag: float = 12.0  #: bright star removal, used for both ZTF filters

    brightPS1gMag: float = 12.0  #: PS1 bright star removal
    brightPS1rMag: float = 12.0  #: PS1 bright star removal
    brightPS1iMag: float = 12.0  #: PS1 bright star removal
    brightPS1zMag: float = 12.0  #: PS1 bright star removal

    removeNegative: int = 1  #: remove negative detections from lightcurve (or not)

    minGalLat: float = 5.0  #: you may stay away from the Galactic plane

    doGaiaVeto: int = 1  #: triggers the use of the Gaia-based filter for the sgsore, designed to remove star that get a high score

    maxPS1matches: int = 100  #: remove source in dense fields (but some galaxies also get many matches, so leave this large!)

    lastOnly: int = (
        0  #: apply all cuts only to the most recent detection in the light curve
    )

    """
    Return values:

    True: Passed all cuts
    
    -10: Large sgscore
    -11: Large sgscore2
    -12: No sgscore available

    -20: Bad PS1 photometry
    -21: PS1 too far
    -22: Bright 1st PS1 match
    -23: Bright 2nd PS1 match
    -24: Birght 3rd PS1 match

    -30: Nothing passed default filter

    -40: Not enough detections

    -50: Too bright in ZTF reference image

    -60: Too many PS1 matches (field too crowded)
    -61: Too close to Galactic Plane (field too crowded)

    -70: RB too low

    -80: Only duplicate detections

    -90: Potential mover

    -100: Large gap between detections

    -110: Too faint

    -120: Flux increase too small

    -130: Gaia veto

    -140: Distance veto
    -141: No distance could be computed

    -150: Last detection did not pass default filter (only when lastOnly=True)
    """

    def post_init(self) -> None:

        # collect bright cuts into dict
        self.brightPS1Mag = {}
        self.brightPS1Mag["g"] = self.brightPS1gMag
        self.brightPS1Mag["r"] = self.brightPS1rMag
        self.brightPS1Mag["i"] = self.brightPS1iMag
        self.brightPS1Mag["z"] = self.brightPS1zMag

        # Instance dictionaries later used in method apply

        # remove upper limits
        isdetect_flt = {"attribute": "candid", "operator": "is not", "value": None}

        # no bad pixels
        nbad_flt = {"attribute": "nbad", "operator": "<=", "value": 0}

        # similar to Growth Marshal mandatory filter (Full Width Half Max assuming a Gaussian core, from SExtractor)
        fwhm_flt = {"attribute": "fwhm", "operator": "<", "value": 5}

        # similar to Growth Marshal mandatory filter (Ratio: aimage / bimage)
        elong_flt = {"attribute": "elong", "operator": "<", "value": 1.3}

        # similar to Growth Marshal mandatory filter (Difference: magap - magpsf )
        magdiff_flt_lower = {"attribute": "magdiff", "operator": ">", "value": -0.3}

        magdiff_flt_upper = {"attribute": "magdiff", "operator": "<", "value": 0.3}

        # has host galaxy detected (removes orphans)
        distnr_flt = {"attribute": "distnr", "operator": ">", "value": 0}

        # combine the current filters
        self._default_filters: List[Dict[str, Any]] = [
            isdetect_flt,
            distnr_flt,
            nbad_flt,
            fwhm_flt,
            elong_flt,
            magdiff_flt_lower,
            magdiff_flt_upper,
        ]

        # remove negative detection (or not)
        if self.removeNegative:

            # is not ref-science
            isdiff_flt1 = {"attribute": "isdiffpos", "operator": "!=", "value": "0"}

            # is not ref-science (again!)
            isdiff_flt2 = {"attribute": "isdiffpos", "operator": "!=", "value": "f"}

            self._default_filters += [isdiff_flt1, isdiff_flt2]

        self.logger.info("NuclearFilter initialized")

    def wmean(self, x, w):
        """
        compute a weighted mean, used for offset statistics
        """
        return np.sum(x * w) / np.sum(w)

    def process(self, alert: AmpelAlertProtocol) -> Optional[Union[bool, int]]:
        """
        Mandatory implementation.
        To exclude the alert, return *None*
        To accept it, either return
        * self.on_match_t2_units
        * or a custom combination of T2 unit names

        Returns None (reject) or True (accept)

        Make a selection on:
        - the distance between the transient and host in reference image
        - the Real/Bogus sore
        - the distance to a bright star

        """

        # first check we have an extended source (note this can remove flares from faint galaxies that missclassified in PS1)
        # these will have to be dealt with in the orphan/faint filter

        if len(alert.get_values("sgscore1")) == 1:

            alert_pps = alert.datapoints[0]
            sgscore1, sgscore2, sgscore3 = (
                alert_pps["sgscore1"],
                alert_pps["sgscore2"],
                alert_pps["sgscore3"],
            )
            distpsnr1, distpsnr2, distpsnr3 = (
                alert_pps["distpsnr1"],
                alert_pps["distpsnr2"],
                alert_pps["distpsnr3"],
            )

        # check that we dont have very old (pre v1.8) schema
        else:

            self.why = "this schema doesnt have sgscore1, rejecting this alert"
            self.logger.info(None, extra={"sgscore1": None})
            return -12

        # print a warning if we have PS1 match issues
        if distpsnr2 < self.closePS1dist:
            self.logger.info(
                "Potential problem with PS1, we have close matches",
                extra={"distpsnr1": distpsnr1, "distpsnr2": distpsnr2},
            )

        # check sgscore, allowing for close PS1 matches with a safescore to overwrite the nearest score
        if (sgscore1 > self.maxSgscore) and not (
            (sgscore2 < self.minSafeSgscore) and (abs(distpsnr2) < self.closePS1dist)
        ):

            self.why = "sgscore1 = {0:0.2f}, which is > {1:0.2f}".format(
                sgscore1, self.maxSgscore
            )
            logextra = {"sgscore1": sgscore1, "maxSgscore": self.maxSgscore}
            if distpsnr2 < self.closePS1dist:
                self.why += "; sgscore2 = {0:0.2f} (dist={1:0.2f})".format(
                    sgscore2, distpsnr2
                )
                logextra["sgscore2"] = sgscore2
                logextra["distpsnr2"] = distpsnr2

            self.logger.info("Large sgscore", extra=logextra)
            return -10

        # also check the sgscore of the close match
        if (sgscore2 > self.maxSgscore) and (abs(distpsnr2) < self.closePS1dist):
            self.why = (
                "sgscore2 = {0:0.2f}, which is > {1:0.2f}  (distpsnr2={2:0.2f})".format(
                    sgscore2, self.maxSgscore, distpsnr2
                )
            )
            self.logger.info(
                "Large sgscore2", extra={"sgscore2": sgscore2, "distpsnr2": distpsnr2}
            )
            return -11

        # check problems with photometry (likely for saturated stars)

        # collect all PS1 magnitude into a dict
        psmag1 = {k: alert_pps["s" + k + "mag1"] for k in ("g", "r", "i", "z")}
        psmag2 = {k: alert_pps["s" + k + "mag2"] for k in ("g", "r", "i", "z")}
        psmag3 = {k: alert_pps["s" + k + "mag3"] for k in ("g", "r", "i", "z")}

        if (psmag1["r"] < 0) and not (
            (psmag2["r"] > 0) and (abs(distpsnr2) < self.closePS1dist)
        ):
            self.why = "1st PS1 match r-band photometry is faulty: mag={0:0.2f} (dist={1:0.2f})".format(
                psmag1["r"], distpsnr1
            )
            if distpsnr2 < self.closePS1dist:
                self.why += "; 2nd PS match r-band also faulty: mag={0:0.2f} (dist={1:0.2f})".format(
                    psmag1["r"], distpsnr1
                )
            self.logger.info(
                "Bad PS1 photometry",
                extra={
                    "psmag1r": psmag1["r"],
                    "distpsnr1": distpsnr1,
                    "psmag2r": psmag2["r"],
                    "distpsnr2": distpsnr2,
                },
            )
            return -20

        # check that the nearest PS1 source is not too far
        if abs(distpsnr1) > self.maxDistPS1source:
            self.why = (
                "distance to 1st PS match is {0:0.2f}, which is > {1:0.2f}".format(
                    distpsnr1, self.maxDistPS1source
                )
            )
            self.logger.info("PS too far", extra={"distpsnr1": distpsnr1})
            return -21

        # check for nearby bright stars
        for k in ("g", "r", "i", "z"):

            if abs(psmag1[k]) < self.brightPS1Mag[k]:
                self.why = "1st PS match in {0}-band, mag={1:0.2f}, which is < {2:0.2f} (dist={3:0.2f})".format(
                    k, psmag1[k], self.brightPS1Mag[k], distpsnr1
                )
                self.logger.info(
                    "Bright 1st PS match",
                    extra={"filter": k, "psmag1": psmag1[k], "distpsnr1": distpsnr1},
                )
                return -22

            if psmag2 is not None:  # old schema check
                if (abs(psmag2[k]) < self.brightPS1Mag[k]) and (
                    abs(distpsnr2) < self.brightObjDist
                ):
                    self.why = "2nd PS match in {0}-band, mag={1:0.2f}, which is < {2:0.2f} (dist={3:0.2f})".format(
                        k, psmag2[k], self.brightPS1Mag[k], distpsnr2
                    )
                    self.logger.info(
                        "Bright 2nd PS match",
                        extra={
                            "filter": k,
                            "psmag2": psmag2[k],
                            "distpsnr2": distpsnr2,
                        },
                    )
                    return -23

            if psmag3 is not None:
                if (abs(psmag3[k]) < self.brightPS1Mag[k]) and (
                    abs(distpsnr3) < self.brightObjDist
                ):
                    self.why = "3rd  PS1 match in {0}-band, mag={1:0.2f}, which is < {2:0.2f} (dist={3:0.2f})".format(
                        k, psmag3[k], self.brightPS1Mag[k], distpsnr3
                    )
                    self.logger.info(
                        "Bright 3rd PS match",
                        extra={
                            "filter": k,
                            "psmag3": psmag3[k],
                            "distpsnr3": distpsnr3,
                        },
                    )
                    return -24

            # don't use the code below because it will remove sources next to objects
            # that were detected in just one pan-starrs band and thus have srmag=-999
            #
            # if ((psmag2['r']<0) or (psmag['g']<0)) and (abs(distpsnr2)< self.brightObjDist):
            #   self.why = "2nd PS1 match saturated(?) sgmag={0:0.2f} srmag={1:0.2f} (dist={2:0.2f})".format(psmag2['g'], psmag2['r'], distpsnr2)
            #   self.logger.info(self.why)
            #   return 0

        # get RealBogus scores for observations, check number of bad pixels
        res_tuples = alert.get_ntuples(
            ["rb", "jd", "magnr"], filters=self._default_filters
        )

        # check that we have anything
        if len(res_tuples) == 0:
            self.why = "nothing passed default filter"
            self.logger.info(self.why)
            return -30

        res = list(map(list, zip(*res_tuples)))
        rb_arr, jd_arr, magnr_arr = np.asarray(res)

        # check number of detections in all bands
        if len(jd_arr) < self.minDetections:
            self.why = "only {0} detections pass default filter, but we require at least {1}".format(
                len(jd_arr), self.minDetections
            )
            self.logger.info("not enough detections", extra={"detections": len(jd_arr)})

            return -40

        minmag = np.min(magnr_arr).item()
        # check that source is not too bright in ZTF ref img
        if self.brightRefMag > minmag:
            self.why = f"min(magnr)={minmag:0.2f}, which is < {self.brightRefMag:0.1f}"
            self.logger.info("too bright in ZTF ref img", extra={"minmag": minmag})

            return -50

        # check source density --> note, we also check Gaia source density, because this doesn't work well for big galaxies
        nmtchps = alert.get_values("nmtchps")[0]
        if nmtchps > self.maxPS1matches:
            self.why = "nmtchps={0}, which is > {1}".format(nmtchps, self.maxPS1matches)
            self.logger.info("nmtchps too large", extra={"nmtchps": nmtchps})
            return -60

        # check Galactic Latitude
        ra, dec = alert.get_values("ra")[0], alert.get_values("dec")[0]

        if ra > -99:
            gc = astropy.coordinates.SkyCoord(
                ra=ra * astropy.units.deg, dec=dec * astropy.units.deg
            ).galactic
            gal_l, gal_b = gc.galactic.l.deg, gc.galactic.b.deg
            if abs(gal_b) < self.minGalLat:
                self.logger.info(
                    "too close to Galactic plane",
                    extra={"gal_l": gal_l, "gal_b": gal_b},
                )

                return -61
        else:
            self.logger.warn(
                "problems with alert coordindates", extra={"ra": ra, "dec": dec}
            )

        # if we want, only check last observation
        if self.lastOnly:

            lastcheck = alert.get_values(
                "jd",
                filters=self._default_filters
                + [
                    {
                        "attribute": "jd",
                        "operator": "==",
                        "value": max(alert.get_values("jd")),
                    }
                ],
            )

            if len(lastcheck) == 0:
                self.why = "last detection did not pass default filter"
                self.logger.info(self.why)
                return -150

            rb_arr = [
                rb_arr[np.argmax(jd_arr)]
            ]  # make sure rb check below is only for last detection

        # if no detections pass real bogus, remove
        if max(rb_arr) < self.minRealBogusScore:
            self.why = "max(rb)={0:0.2f}, which is  < {1:0.2f}".format(
                max(rb_arr), self.minRealBogusScore
            )
            self.logger.info("max(rb) too low", extra={"max_rb": max(rb_arr)})

            return -70

        # do cut on moving sources (with all good detections)
        dt = abs(np.sort(jd_arr)[1:] - np.sort(jd_arr)[0:-1])

        # first check that we dont have bunch of duplicates
        if not sum(dt > 0):
            self.why = "number of detections={0}, but time difference between all is zero".format(
                len(dt)
            )
            self.logger.info(
                "only duplicate detections", extra={"len(dt)": int(len(dt))}
            )

            return -80

        # check that time between detection is not too short (as expected for moving object)
        dt = dt[dt > 0]
        if np.max(dt) < self.minDeltaJD:
            self.why = "potential mover, number of good detections={0}; \
                        max(time diff)={1:1.3f} h, which is < {2:0.3f} h".format(
                len(jd_arr), max(dt) * 24, self.minDeltaJD * 24
            )
            self.logger.info(
                "potential mover",
                extra={"len_jd_arr": len(jd_arr), "max_dt_hour": max(dt) * 24},
            )

            return -90

        # Check time between detecetions (this could be made more fancy using upper limits etc.)
        if np.min(dt) > self.maxDeltaJD:
            self.why = "number of good detections={0}; min(time diff)={1:1.1f} h, \
                        which is > {2:0.1f} h".format(
                len(jd_arr), min(dt) * 24, self.maxDeltaJD * 24
            )
            self.logger.info(
                "large gap between detection",
                extra={"len_jd_arr": len(jd_arr), "min_dt_hour": min(dt) * 24},
            )

            return -100

        # get some more arrays
        res = list(
            map(
                list,
                zip(
                    *alert.get_ntuples(
                        ["distnr", "magpsf", "sigmapsf", "rb", "fwhm", "fid"],
                        filters=self._default_filters,
                    )
                ),
            )
        )

        distnr_arr, magpsf_arr, sigmapsf_arr, rb_arr, fwhm_arr, fid_arr = np.asarray(
            res
        )

        # get indices to bands
        idx_g = fid_arr == 1
        idx_r = fid_arr == 2

        # removed check on detections per filter in Feb 2019; it's too strict, we can deal with this at the light curve fitting stage
        # if (sum(idx_g)<self.minDetections) and (sum(idx_r)<self.minDetections):
        #   self.why = "number of good detections in (g, r)=({0}, {1}), which is < {2:0}".format(sum(idx_g), sum(idx_r), self.minDetections)
        #   self.logger.info("not enough good detection",
        #       extra={"sum_idx_g":int(sum(idx_g)), "sum_idx_r":int(sum(idx_r))}
        #   )

        #   return 0

        # apply magnitude cut to brightest detection
        if np.min(magpsf_arr) > self.diffmagLimit:
            self.why = "too faint, min(mag)={0:0.2f}, which is \
                        > {1:0.1f}".format(
                np.min(magpsf_arr), self.diffmagLimit
            )
            self.logger.info("too faint", extra={"min_magpsf_arr": np.min(magpsf_arr)})

            return -110

        # apply lower limit on flux increase
        if np.min(magpsf_arr - magnr_arr) > self.maxDeltaMag:
            self.why = "flux increase too small, min(magpsf-magnr)={0:0.2f}, \
                        which is > {1:0.2f}".format(
                np.min(magpsf_arr - magnr_arr), self.maxDeltaMag
            )
            self.logger.info(
                "flux increase too small",
                extra={"min_magpsfnr_diff": np.min(magpsf_arr - magnr_arr)},
            )
            return -120

        # allow Gaia to veto
        if self.doGaiaVeto:
            G_veto = self.gaia_veto(alert)
            if G_veto:
                return -130

        # if we make it this far, compute the host-flare distance, using only (decent-enough) detections
        res = list(
            map(
                list,
                zip(
                    *alert.get_ntuples(
                        ["ra", "dec", "ranr", "decnr"], filters=self._default_filters
                    )
                ),
            )
        )

        ra_arr, dec_arr, ranr_arr, decnr_arr = np.asarray(res)

        # compute a few different measures of the distance
        # we compute these for each band seperately
        sigma_offset = np.clip(
            0.24 + 0.04 * (magpsf_arr - 20), 0.1, 1
        )  # equation from the NedStark paper
        snr_weight = 1 / sigma_offset**2

        offset_info = {}

        for idx, bnd in zip([idx_g, idx_r], ["g", "r"]):

            if sum(idx) >= self.minDetections:

                mean_dist = 3600 * np.sqrt(
                    np.mean((ra_arr - ranr_arr)[idx]) ** 2
                    + np.mean((dec_arr - decnr_arr)[idx]) ** 2
                )
                median_dist = 3600 * np.sqrt(
                    np.median((ra_arr - ranr_arr)[idx]) ** 2
                    + np.median((dec_arr - decnr_arr)[idx]) ** 2
                )
                weighted_dist = 3600 * np.sqrt(
                    self.wmean((ra_arr - ranr_arr)[idx], snr_weight[idx]) ** 2
                    + self.wmean((dec_arr - decnr_arr)[idx], snr_weight[idx]) ** 2
                )

                offset_info[bnd] = (mean_dist, median_dist, weighted_dist)

                if (mean_dist < self.maxDeltaRad) and (mean_dist > self.minDeltaRad):
                    self.why = "pass on mean dist={0:0.2f}, band={1}; detections used={2}".format(
                        mean_dist, bnd, sum(idx)
                    )

                    self.logger.info(
                        "pass on mean dist",
                        extra={
                            "mean_dist": mean_dist,
                            "band": bnd,
                            "detections": int(sum(idx)),
                        },
                    )
                    return True

                if (median_dist < self.maxDeltaRad) and (
                    median_dist > self.minDeltaRad
                ):
                    self.why = "pass on median dist={0:0.2f}, band={1}; detections used={2}".format(
                        median_dist, bnd, sum(idx)
                    )

                    self.logger.info(
                        "pass on median dist",
                        extra={
                            "medianDist": median_dist,
                            "band": bnd,
                            "detections": int(sum(idx)),
                        },
                    )

                    return True

                if (weighted_dist < self.maxDeltaRad) and (
                    weighted_dist > self.minDeltaRad
                ):
                    self.why = "pass on weighted dist={0:0.2f}, band={1}; detections used={2}".format(
                        weighted_dist, bnd, sum(idx)
                    )
                    self.logger.info(
                        "pass on weighted dist",
                        extra={
                            "weightedDist": weighted_dist,
                            "band": bnd,
                            "detections": int(sum(idx)),
                        },
                    )

                    return True

        # if none of the measures of the host-flare pass the cut, reject this alert
        if len(offset_info.keys()):
            self.why = ""
            for bnd in offset_info.keys():
                self.why += "mean/median/weighted dist in {0}-band = ({1:0.2f}/{2:0.2f}/{3:0.2f}), which is > {4:0.2f}".format(
                    bnd,
                    offset_info[bnd][0],
                    offset_info[bnd][1],
                    offset_info[bnd][2],
                    self.maxDeltaRad,
                )

                self.logger.info(
                    None,
                    extra={
                        "meanDist-" + bnd: mean_dist,
                        "weightedDist-" + bnd: weighted_dist,
                        "medianDist-" + bnd: median_dist,
                        "band": bnd,
                    },
                )

            return -140
        else:
            self.why = "offset distance not created, likely not enough data passed default filter"
            self.logger.info(self.why, extra={"sumIdx": int(sum(idx))})
            return -141



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/nuclear/t0/GaiaVeto.py
# License           : BSD-3-Clause
# Author            : sjoertvv <sjoert@astro.umd.edu>
# Date              : 26.02.2018
# Last Modified Date: 05.04.2023
# Last Modified By  : simeon.reusch@desy.de

from typing import Any, Optional, Sequence, Union

import numpy as np
import pandas as pd  # type: ignore

from ampel.protocol.AmpelAlertProtocol import AmpelAlertProtocol
from ampel.ztf.base.CatalogMatchUnit import CatalogMatchUnit


class GaiaVetoMixin(CatalogMatchUnit):
    maxGaiamatches: int = 30  #: remove source in dens fields (almost certainly star)
    brightObjDist: float = (
        15.0  #: max distance in arcsec for checking nearby bright stars in Gaia or PS1
    )
    brightGaiaMag: float = 13.5  #: Gaia bright star removal in mulitple bands
    maxParallaxSNR: float = 3.0  #: Gaia parallax cut

    singleGaiaDist: float = (
        1.0  #: Distance to check multiple matches in Gaia (ie, reject double stars)
    )

    minDeltaG: float = 0.2  #: Difference between Gaia mag and PS1_converted_Gaia (large values imply more flux in PS1, ie extended sources) only applied to source with sgscore1 >= minSgscore

    minSafeSgscore: float = (
        0.5  # min star-galaxy score for which we dont allow a Gaia-based veto
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # conversion from PS1 to SDSS (https://arxiv.org/abs/1203.0297)
        flt = ("g", "r", "i", "z")
        self.__A0 = {flt[i]: x for i, x in enumerate((0.013, -0.001, -0.005))}
        self.__A1 = {flt[i]: x for i, x in enumerate((0.145, 0.004, 0.011))}
        self.__A2 = {flt[i]: x for i, x in enumerate((0.019, 0.007, 0.010))}

    def alert_to_Gaia(
        self, alert: AmpelAlertProtocol, force_ztf=False
    ) -> Optional[float]:
        """
        convert from PS1/ZTF to Gaia (used in Gaia veto function)
        """

        # get the PS1 PSF photometry
        ps1_mag = {k: alert.get_values("s" + k + "mag1")[0] for k in ("g", "r", "i")}

        # if we have a problem with the PS1 photometry, try to use ZTF
        if (min([ps1_mag[k] for k in ps1_mag]) < 0) or force_ztf:
            if not force_ztf:
                self.logger.warn(
                    "no decent PS1 photometry ({0:0.2f} {1:0.2f} {2:0.2f}), trying to use ZTF to estimate Gaia flux".format(
                        ps1_mag["g"], ps1_mag["r"], ps1_mag["i"]
                    )
                )

            magnr_g = alert.get_values(
                "magnr", filters=[{"attribute": "fid", "operator": "==", "value": 1}]
            )
            magnr_R = alert.get_values(
                "magnr", filters=[{"attribute": "fid", "operator": "==", "value": 2}]
            )

            if len(magnr_g) and len(magnr_R):
                ztf_g, ztf_r = np.median(magnr_g), np.median(magnr_R)
                gr = ztf_g - ztf_r

                # Gaia_from_ztf = ztf_g -0.0662 -0.7854*gr -0.2859*gr**2 +0.0145*gr**3          # from Jordi (2010) #https://arxiv.org/abs/1008.0815
                Gaia_from_ztf = (
                    ztf_g
                    - 0.038025
                    - 0.76988 * gr
                    - 0.1931 * gr**2
                    + 0.0060376 * gr**3
                )  # from Evan (2018) #https://arxiv.org/abs/1804.09368
                return Gaia_from_ztf
            else:
                self.logger.warn(
                    "not enough ZTF detections (g={0}, R={1}) to compute Gaia mag".format(
                        len(magnr_g), len(magnr_R)
                    )
                )
                return None

        else:
            # convert from PS1 to SDSS (https://arxiv.org/abs/1203.0297)
            gr_ps1 = ps1_mag["g"] - ps1_mag["r"]
            sdss_mag = {
                flt: ps1_mag[flt]
                + self.__A0[flt]
                + self.__A1[flt] * (gr_ps1)
                + self.__A2[flt] * (gr_ps1) ** 2
                for flt in ("g", "i")
            }

            # convert from SDSS to Gaia
            gi = sdss_mag["g"] - sdss_mag["i"]
            # Gaia_from_sdss = sdss_mag['g']-0.0912   -0.5310*gi  -0.1042*gi**2   +0.0068*gi**3     # (DR0, Jordi+18, https://arxiv.org/abs/1008.0815)
            Gaia_from_sdss = (
                sdss_mag["g"]
                - 0.074189
                - 0.51409 * gi
                - 0.080607 * gi**2
                + 0.0016001 * gi**3
            )  # (DR2, Evans+18, https://arxiv.org/abs/1804.09368, v3)
            return Gaia_from_sdss

    # no useful type hints for astropy.Table
    def get_Gaia(self, alert: AmpelAlertProtocol) -> Optional[pd.DataFrame]:
        """
        function to run catsHTM.cone_search and convert to np.array
        """
        ra, dec = np.median(alert.get_values("ranr")), np.median(
            alert.get_values("decnr")
        )

        # some alerts contain -999 values, for whatever reason
        if ra < 0 or dec < -90 or dec > 90:
            return None

        matches = self.cone_search_all(
            ra,
            dec,
            [
                {
                    "name": "GAIADR2",
                    "use": "catsHTM",
                    "rs_arcsec": 30,
                    "keys_to_append": ["Mag_G", "Plx", "ErrPlx"],
                }
            ],
        )[0]

        if not matches:
            return None

        return pd.DataFrame(
            [{"dist_arcsec": m["dist_arcsec"], **m["body"]} for m in matches]
        )

    def gaia_veto(self, alert: AmpelAlertProtocol) -> bool:
        """
        veto IPAC star/galaxy score based on Gaia:
         - parallax
         - too many close matches
         - too many matches
         - compactness: Gaia flux compared to PS1 (or ZTF) flux
        """

        if (gaia_df := self.get_Gaia(alert)) is None:
            return False

        if len(gaia_df) > self.maxGaiamatches:
            self.why = (
                """number of Gaia matches within 30" is {0}, which is > {1}""".format(
                    len(gaia_df), self.maxGaiamatches
                )
            )
            self.logger.info(
                '''too many Gaia matches within 30"''',
                extra={"n_Gaia_matches_within_30ac": len(gaia_df)},
            )
            return True

        idxBright = gaia_df["dist_arcsec"] < self.brightObjDist
        if sum(idxBright):
            gaia_brighest_match = np.min(gaia_df["Mag_G"][idxBright])
        else:
            self.logger.info(
                """no Gaia matches within {0:0.1f}" (not vetoing)""".format(
                    self.brightObjDist
                )
            )
            return False

        if gaia_brighest_match < self.brightGaiaMag:
            self.why = """within {0:0.1f}" of bright star in Gaia: G={1:0.2f}, which is {2:0.1f}""".format(
                self.brightObjDist, gaia_brighest_match, self.brightGaiaMag
            )
            self.logger.info(
                "close to Gaia bright star",
                extra={"gaia_brighest_match": gaia_brighest_match},
            )
            return True

        idxNear = gaia_df["dist_arcsec"] < self.singleGaiaDist
        if sum(idxNear) > 1:
            self.why = """number of Gaia matches within {0:0.1f}" is {1}, which is one too many""".format(
                self.singleGaiaDist, sum(idxNear)
            )
            self.logger.info(
                "Close Gaia match", extra={"n_close_Gaia_matches": int(sum(idxNear))}
            )
            return True
        elif sum(idxNear) == 0:
            self.logger.info("no Gaia match (not vetoing)")
            return False

        # with one left, we can look at the nearest match only
        gaiam = gaia_df[idxNear].iloc[0]

        if (
            gaiam["Plx"] is not None
            and gaiam["ErrPlx"] is not None
            and (parallax_snr := gaiam["Plx"] / gaiam["ErrPlx"]) > self.maxParallaxSNR
        ):
            self.why = """parallax SNR={0:0.1f}, which is > {1:0.1f}""".format(
                parallax_snr, self.maxParallaxSNR
            )
            self.logger.info(
                "gaia parallax, vetoing", extra={"parallax_snr": parallax_snr}
            )
            return True

        # if the score is in the "safe" range, don't allow veto based on photometry
        sg1 = alert.get_values("sgscore1")[0]
        if sg1 < self.minSafeSgscore:
            self.logger.info(
                "not vetoing this alert because sgscore={0:0.2f}, which is < {1:0.2f}".format(
                    sg1, self.minSafeSgscore
                )
            )
            return False

        # if not in safe range, compute the difference between Gaia and ground-based photometry (PS1/ZTF)
        predicted_G = self.alert_to_Gaia(alert)
        if predicted_G is not None:
            delta_G = gaiam["Mag_G"] - predicted_G
            self.why = "Gaia - Ground = {0:0.2f}".format(delta_G)
            self.logger.info(None, extra={"delta_G": delta_G})
        else:
            self.why = """Can't compute Gaia magnitude from PS1 or ZTF data, vetoing"""
            self.logger.info(self.why)
            return True

        if delta_G < self.minDeltaG:
            self.why = """Gaia - Ground = {0:0.2f}, which is < {1:0.2f}""".format(
                delta_G, self.minDeltaG
            )
            self.logger.info("too compact in Gaia, vetoing", extra={"delta_G": delta_G})
            return True

        return False
